import upswingutil as ul
from upswingutil.db import MongodbV2
from upswingutil.db.model import ReservationModel, Status


ul.MONGO_URI = 'mongodb://AdminUpSwingGlobal:Upswing098812Admin0165r@dev.db.upswing.global:27017/?authSource=admin&readPreference=primary&appname=Agent%20Oracle%20Dev&ssl=false'


if __name__ == '__main__':
    print('testing resv creation')
    orgId = 'OHIPB2'
    mongo = MongodbV2(orgId)
    resv = ReservationModel(_id="t2")
    resv.agent = 'test'
    resv.status = Status.CANCELLED
    mongo.save(resv)
    mongo.close_connection()
    print('completed')



from enum import Enum

from mongoengine import *
from upswingutil.db.model import ProfileIDObject, GuestsModel


class Status(Enum):
    NO_SHOW = 'NoShow'
    DEPARTED = 'Departed'
    RESERVED = 'Reserved'
    CANCELLED = 'Cancelled'
    ARRIVED = 'Arrived'
    CONFIRMED = 'Confirmed'
    CREATED = 'Created'
    UNCONFIRMED = 'Unconfirmed'
    MAINTENANCE = 'Maintenance'
    IN_HOUSE = 'InHouse'
    PENCIL = 'Pencil'


class ReservationIDObj(DynamicEmbeddedDocument):
    reservation = StringField()
    confirmation = StringField()
    globalId = StringField()


class AlvieReservationModel(DynamicEmbeddedDocument):
    comments = ListField(StringField())
    rating = DictField()


class ReservationGuestShort(DynamicEmbeddedDocument):
    idObj = EmbeddedDocumentField(ProfileIDObject)
    primary = BooleanField()
    arrivalTransport = DictField()
    departureTransport = DictField()
    # birthDate = StringField()
    guest = ReferenceField(GuestsModel)


class ReservationGuestInfo(DynamicEmbeddedDocument):
    adults = IntField()
    children = IntField()
    infants = IntField()
    childBuckets = DictField()
    preRegistered = BooleanField()
    guest_list = EmbeddedDocumentListField(ReservationGuestShort)


class ReservationModel(DynamicDocument):
    meta = {'collection': 'reservations'}

    _id = StringField(required=True, primary_key=True)
    agent = StringField(required=True)
    alvie = EmbeddedDocumentField(AlvieReservationModel)
    arrivalDate = StringField(required=True)
    bookingInfo = DictField()
    historyEvents = ListField(DictField())
    cancellation = ListField(DictField())
    comments = ListField(DictField())
    createBusinessDate = StringField()
    createDateTime = StringField()
    daily_activity = ListField(DictField())
    departureDate = StringField()
    eCertificates = ListField(DictField())
    expectedTimes = DictField()
    financeInfo = DictField()
    folioInformation = DictField()
    guestInfo = EmbeddedDocumentField(ReservationGuestInfo)
    guestLocators = DictField()
    hotelId = StringField(required=True)
    hotelName = StringField(required=True)
    idObj = EmbeddedDocumentField(ReservationIDObj)
    inventoryItems = DictField()
    linkedReservation = DictField()
    memberships = ListField(DictField())
    metaInfo = DictField()
    orgId = StringField()
    originalTimeSpan = DictField()
    packages = ListField(DictField())
    policies = DictField()
    preferences = ListField(DictField())
    housekeeping = ListField(DictField())
    roomStay = DictField()
    status = EnumField(Status, default=Status.CREATED, required=True)
    auraRecordUpdatedOn = StringField()


class ReservationModelV2(DynamicDocument):
    meta = {'collection': 'reservations'}

    _id = StringField(required=True, primary_key=True)
    orgId = StringField()
    idObj = EmbeddedDocumentField(ReservationIDObj)
    agent = StringField(required=True)
    datesAndDuration = DictField()
    status = EnumField(Status, default=Status.CREATED, required=True)
    correspondence = ListField(DictField())
    metaInfo = DictField()
    property = DictField(required=True)
    stayInfo = DictField(required=True)
    reservationTypes = DictField(required=True)
    bookingInfo = DictField(required=True)
    financeInfo = DictField(required=True)
    daily_activity = ListField(DictField())
    linkedReservation = ListField(DictField())
    guestInfo = EmbeddedDocumentField(ReservationGuestInfo)
    eCertificates = ListField(DictField())
    historyEvents = ListField(DictField())
    cancellation = ListField(DictField())
    housekeeping = ListField(DictField())
    comments = ListField(DictField())
    policies = DictField()
    inventoryItems = DictField()
    preferences = ListField(DictField())
    requirement = ListField(DictField())
    rego_access = ListField(DictField())
    memberships = ListField(DictField())
    packages = ListField(DictField())
    transfers = ListField(DictField())
    event = DictField()
    alvie = EmbeddedDocumentField(AlvieReservationModel)
    auraRecordUpdatedOn = DateTimeField()



import datetime
import logging
import unittest
import firebase_admin
from firebase_admin import credentials

import upswingutil as ul
from upswingutil.pms import rms, oracle
from upswingutil.resource import access_secret_version, setup_logging
import json


class TestReservationV2API(unittest.TestCase):
    ul.G_CLOUD_PROJECT = 'aura-staging-31cae'
    ul.ENCRYPTION_SECRET = "S1335HwpKYqEk9CM0I2hFX3oXa5T2oU86OXgMSW4s6U="
    ul.MONGO_URI = access_secret_version(ul.G_CLOUD_PROJECT, 'MONGOURI', '5')
    ul.LOG_LEVEL_VALUE = logging.DEBUG

    def test_adding_new_RMS_reservation(self):
        setup_logging()
        print('Initializing default firebase app')
        cred = json.loads(
            access_secret_version(ul.G_CLOUD_PROJECT, 'Firebase-Aura', '1'))
        cred = credentials.Certificate(cred)
        firebase_admin.initialize_app(cred)

        print('Initializing alvie firebase app')
        cred = json.loads(
            access_secret_version(ul.G_CLOUD_PROJECT, 'Firebase-Alvie', '3'))
        cred = credentials.Certificate(cred)
        firebase_admin.initialize_app(cred, name='alvie')

        resv = rms.ReservationSync('11281')
        resv_list = [204862]
        for _resv_id in resv_list:
            start_time = datetime.datetime.now()
            print(f'processing reservation : {_resv_id}')
            resv.process(_resv_id)
            print(f'processed reservation : {_resv_id}')
            end_time = datetime.datetime.now()
            print(f'total time to process : {end_time-start_time}')

    def test_adding_new_ORACLE_reservation(self):
        setup_logging()
        print('Initializing default firebase app')
        cred = json.loads(
            access_secret_version(ul.G_CLOUD_PROJECT, 'Firebase-Aura', '1'))
        cred = credentials.Certificate(cred)
        firebase_admin.initialize_app(cred)

        print('Initializing alvie firebase app')
        cred = json.loads(
            access_secret_version(ul.G_CLOUD_PROJECT, 'Firebase-Alvie', '1'))
        cred = credentials.Certificate(cred)
        firebase_admin.initialize_app(cred, name='alvie')

        resv = oracle.ReservationSync('OHIPSB')
        resv_list = [
            {
                'reservation': 17151,
                'hotelId': 'SAND01'
            },
            {
                'reservation': 16960,
                'hotelId': 'SAND01'
            },
            {
                'reservation': 17152,
                'hotelId': 'SAND01'
            },
            {
                'reservation': 17406,
                'hotelId': 'SAND01'
            },
            {
                'reservation': 17400,
                'hotelId': 'SAND01'
            },
            {
                'reservation': 16995,
                'hotelId': 'SAND01'
            },
            {
                'reservation': 17007,
                'hotelId': 'SAND01'
            },
        ]
        for _resv in resv_list:
            print(f'processing reservation : {_resv}')
            resv.process(_resv.get('reservation'), _resv.get('hotelId'))
            print(f'processed reservation : {_resv}')



import logging
import ssl
import os
import mongoengine
from loguru import logger
import upswingutil as ul

__PII_FIELDS__ = ['guestGiven', 'guestSurname', 'mobile', 'passportId', 'email', 'email2']

mongoengine.connect(host='mongodb://ReadWrite:gbyate0ewA712Dkq823gswX9i1oP@dev.db.upswing.global:27017/?authSource=admin&readPreference=primary&appname=Aura&ssl=false')


class MongodbV2:
    RESOURCE_COLLECTION = 'resource'
    AREAS_COLLECTION = "areas"
    GUEST_COLLECTION = "guests"
    PROPERTY_COLLECTION = "properties"
    RESERVATION_COLLECTION = "reservations"
    TRANSACTION_COLLECTION = "transactions"
    AREA_REPORTS_BY_DAY_COLLECTION = "area_report_by_day"
    COUNTRIES_COLLECTION = 'countries'
    ORG_CONFIG_COLLECTION = 'org_config'
    COUNTERS_COLLECTION = 'counters'
    EVENTS_COLLECTION = 'events'
    DAILY_REPORTS_COLLECTION = 'daily_reports'

    def __init__(self, db_name):
        try:
            self.db_name = str(db_name)
            mongoengine.get_connection(db_name)
        except mongoengine.connection.ConnectionFailure as err:
            mongoengine.connect(db_name, host=os.getenv('MONGO_URI', ul.MONGO_URI), alias=db_name)
        except Exception as e:
            logger.error('Error while connecting to db')
            logger.error(type(e))

    def save(self, doc: mongoengine.DynamicDocument):
        logger.info(f'Data saved successfully to {self.db_name} : {doc.id}')
        doc.switch_db(self.db_name)
        doc.save()

    def close_connection(self):
        try:
            mongoengine.disconnect(self.db_name)
        except Exception as e:
            logger.error('Error while closing connection to db')
            logger.error(e)


if __name__ == '__main__':
    mongo = MongodbV2('11249')
    mongo.close_connection()



from typing import Optional
from pydantic import BaseModel

class MultiText(BaseModel):
    defaultText: str = ''
    translatedTexts: Optional[list] = []

# class Languages(BaseModel):
#     description: MultiText = MultiText()
#     translationLanguageCode: str
#     reportDateLanguage: str = ''
#     languageCode = str

class CommunicationMethodsEntDetails(BaseModel):
    code: str
    description: MultiText = MultiText()
    displayOrder: int

class HotelTaxType(BaseModel):
    code: str
    description: str
    collectingAgentTax: Optional[bool]
    printAutoAdjust: Optional[bool]
    reportExemptDays: int
    reportTaxPercentage: int


class LeisureActivityTypes(BaseModel):
    typeCode: str
    description: str

class BlockReservationMethodCodes(BaseModel):
    code: str
    description: MultiText = MultiText()
    displayOrder: int





import json
from datetime import datetime

from upswingutil.db.model import Status, ReservationGuestInfo, ReservationGuestShort, \
    ReservationModelV2, GuestsModelV2
from upswingutil.pms import store_reservation_to_alvie
from upswingutil.resource import get_model_from_cloud_storage
from upswingutil.pms.rms import NAME, get_key, validate_key
from upswingutil.db import Mongodb, Firestore, MongodbV2
from upswingutil.schema import Token
import requests
import logging
from pycountry_convert import country_name_to_country_alpha2
from concurrent.futures import ThreadPoolExecutor, as_completed
from upswingutil.pms.rms import COUNTRIES


def _get_status(status):
    if status == 'NoShow':
        return Status.NO_SHOW
    elif status == 'Departed':
        return Status.DEPARTED
    elif status == 'Reserved':
        return Status.RESERVED
    elif status == 'Cancelled':
        return Status.CANCELLED
    elif status == 'Arrived':
        return Status.IN_HOUSE
    elif status == 'Maintenance':
        return Status.MAINTENANCE
    elif status == 'Confirmed':
        return Status.CONFIRMED
    elif status == 'Pencil':
        return Status.PENCIL
    elif status == 'Unconfirmed':
        return Status.UNCONFIRMED
    else:
        logging.error(f'Missing Status : {status}')
        return Status.UNCONFIRMED


def _create_reservation_id_obj(record) -> dict:
    return {
        'reservation': str(record.get('id')),
        'globalId': f"{record.get('agent')}-{record.get('orgId')}-{record.get('clientId')}-{record.get('id')}",
        'accountId': str(record.get('accountId')),
        'activityId': str(record.get('activityId')),
        'boatId': str(record.get('boatId')),
        'bookerContactId': str(record.get('bookerContactId')),
        'businessLostId': str(record.get('businessLostId')),
        'businessSegmentId': str(record.get('businessSegmentId')),
        'cancellationPolicyId': str(record.get('cancellationPolicyId')),
        'cancelledById': str(record.get('cancelledById')),
        'contactId': str(record.get('contactId')),
        'confirmedById': str(record.get('confirmedById')),
        'coordinator1Id': str(record.get('coordinator1Id')),
        'coordinator2Id': str(record.get('coordinator2Id')),
        'createdById': str(record.get('createdById')),
        'destinationCodeId': str(record.get('destinationCodeId')),
        'externalCancelId': str(record.get('externalCancelId')),
        'externalResId': str(record.get('externalResId')),
        'fixedResReasonId': str(record.get('fixedResReasonId')),
        'modifiedById': str(record.get('modifiedById')),
        'mealPlanId': str(record.get('mealPlanId')),
        'occupantId': str(record.get('occupantId')),
        'paymentModeId': str(record.get('paymentModeId')),
        'splitResId': str(record.get('splitResId')),
        'subResTypeId': str(record.get('subResTypeId')),
        'bookingSourceId': str(record.get('bookingSourceId')),
        'companyId': str(record.get('companyId')),
        'discountId': str(record.get('discountId')),
        'groupAllotmentId': str(record.get('groupAllotmentId')),
        'groupReservationId': str(record.get('groupReservationId')),
        'marketSegmentId': str(record.get('marketSegmentId')),
        'onlineConfirmationId': str(record.get('onlineConfirmationId')),
        'resTypeId': str(record.get('resTypeId')),
        'subMarketSegmentId': str(record.get('subMarketSegmentId')),
        'travelAgentId': str(record.get('travelAgentId')),
        'voucherId': str(record.get('voucherId')),
        'wholesalerId': str(record.get('wholesalerId')),
        'roomId': str(record.get('areaId')),
        'hotelId': str(record.get('clientId')),
        'categoryId': str(record.get('categoryId')),
        'rateTypeId': str(record.get('rateTypeId')),
    }


def _create_guest_id_obj(record) -> dict:
    return {
        'profile': str(record.get('id')),
        'corporateId': str(record.get('companyId'))
    }


def _get_geo_region(nationalID: int):
    result = 'unknown'
    try:
        mongo = Mongodb('upswing')
        region = mongo.get_collection('countries').find_one({"_id": nationalID}, {"region": 1})
        mongo.close_connection()
        result = region if region else 'unknown'
    except Exception as e:
        logging.error(f'Error getting geo region of nation {nationalID}')
        logging.error(e)
    finally:
        return result


def _generate_day_to_day_entry(record):
    _daily_activity = list()
    for index, item in enumerate(record.get('daily_rates')):
        _daily_revenue = next((x for x in record.get('daily_revenue')), {})
        _daily_activity.append({
            'date': item['stayDate'].split(' ')[0],
            'day_of_stay': index + 1,
            'roomType': record.get('categoryName'),
            'ratePlanCode': record.get('rateTypeName'),
            'rateTypeId': item.get('rateTypeId'),
            'roomId': record.get('areaId'),
            'roomName': record.get('areaName'),
            'rates': {
                'currency': item.get('currency'),
                'amountBeforeTax': item.get('rateAmount'),
                'exclusiveTax': item.get('exclusiveTax'),
                'discountAmount': item.get('discountAmount'),
                'additionalsAmount': item.get('additionalsAmount'),
                'dynamicAmount': item.get('dynamicAmount'),
                'linkedRateAdjustmentAmount': item.get('linkedRateAdjustmentAmount'),
                'packageAmount': item.get('packageAmount'),
                'xNightsDiscount': item.get('xNightsDiscount'),
                'totalRateAmount': item.get('totalRateAmount')
            },
            'revenue': {
                'accommodation': _daily_revenue.get('accommodation'),
                'accommodationTax': _daily_revenue.get('accommodationTax'),
                'accommodationGST': _daily_revenue.get('accommodationGST'),
                'foodAndBeverage': _daily_revenue.get('foodAndBeverage'),
                'foodAndBeverageTax': _daily_revenue.get('foodAndBeverageTax'),
                'foodAndBeverageGST': _daily_revenue.get('foodAndBeverageGST'),
                'other': _daily_revenue.get('other'),
                'otherTax': _daily_revenue.get('otherTax'),
                'otherGST': _daily_revenue.get('otherGST')
            },
            'guestCounts': {
                'adults': record.get('adults'),
                'children': record.get('children'),
                'infants': record.get('infants')
            }
        })
    return _daily_activity


def _retrieve_data(session, id, name, url):
    result = None
    try:
        logging.debug(f'Extracting {name} for reservation: {id}')
        with session.get(url) as response:
            if response.status_code == 200:
                result = response.json()
            else:
                logging.debug(f'Error getting {name} for reservation {id} due to status code {response.status_code}')
                logging.debug(url.format(id))
    except Exception as e:
        logging.error(f' {name} failed due to {e}')
    finally:
        return name, result


class ReservationSync:

    def __init__(self, orgId: str, g_cloud_token=None):
        self.orgId = orgId
        self.mongo = Mongodb(orgId)
        self.mongoV2 = MongodbV2(orgId)
        self.firestore_db = Firestore(app='alvie')
        self.token: Token = get_key(self.orgId)
        self.g_cloud_token = g_cloud_token
        self.__urgent_booking_criteria__ = 5 * 86400
        self.__stay_duration_criteria__ = 10 * 86400
        self.__booking_category_levels__ = ['bronze', 'silver', 'gold', 'platinum']

    def _get_booking_type(self, arrival_date: str, created_date: str):
        result = 'unknown'
        try:
            delta = datetime.fromisoformat(arrival_date) - datetime.fromisoformat(created_date)
            delta = delta.days * 86400 + delta.seconds
            result = 'Urgent Booking' if delta < self.__urgent_booking_criteria__ else 'Pre Planned'
        except Exception as e:
            logging.error('Error while calculating booking type')
            logging.error(e)
        finally:
            return result

    def _get_duration_type(self, departure_date: str, arrival_date: str):
        result = 'unknown'
        try:
            stay_d = datetime.fromisoformat(departure_date) - datetime.fromisoformat(arrival_date)
            delta = stay_d.days * 86400 + stay_d.seconds
            result = 'Short Stay' if delta < self.__stay_duration_criteria__ else 'Long Stay'
        except Exception as e:
            logging.error('Error calculating stay duration type')
            logging.error(e)
        finally:
            return result

    def _get_booking_level(self, orgId, clientId, departure_date: str, arrival_date: str, spend, booking_type):
        result = 'unknown'
        try:
            stay_d = datetime.fromisoformat(departure_date) - datetime.fromisoformat(arrival_date)
            booking_type_id = 1 if booking_type == 'Urgent Booking' else 0
            model = get_model_from_cloud_storage(
                str(orgId),
                f'booking_categorization_{clientId}.pkl',
                token=self.g_cloud_token
            )
            prediction = model.predict([[stay_d.days, spend, booking_type_id]])
            level = prediction[0] if len(prediction) > 0 else None
            result = self.__booking_category_levels__[level] if level else 'unknown'
        except Exception as e:
            logging.error('Error calculating booking_level of reservation')
            logging.error(e)
        finally:
            return result

    def _transform_guest_profile(self, record):
        countryCode = ''
        nationalityCode = ''
        try:
            countryCode = country_name_to_country_alpha2(
                next((x['name'] for x in COUNTRIES if x['id'] == record.get('countryId')), ''))
        except Exception as e:
            logging.error(f'No countryCode found for {record.get("countryId")}')
            logging.error(e)
        try:
            nationalityCode = country_name_to_country_alpha2(
                next((x['name'] for x in COUNTRIES if x['id'] == record.get('nationalityId')), ''))
        except Exception as e:
            logging.error(f'No nationalityCode found for {record.get("nationalityId")}')
            logging.error(e)

        hotelId = self.mongo.get_collection(self.mongo.PROPERTY_COLLECTION).find_one({'primaryDetails.propertyId': str(record.get('propertyId'))}, {'_id': 1})
        hotelId = hotelId.get('_id') if hotelId else '0'
        guest = GuestsModelV2(
            _id=str(record.get('id')),
            idObj=_create_guest_id_obj(record),
            profileType='Guest',
            title=record.get('title'),
            firstName=record.get('guestGiven'),
            lastName=record.get('guestSurname'),
            datesAndDurations={
                'birthDate': record.get('birthday'),
                'anniversary': None
            },
            address={
                'addressLine1': record.get('addressLine1'),
                'addressLine2': record.get('addressLine2'),
                'addressLine3': record.get('addressLine3'),
                'countryId': countryCode,
                'nationalityId': nationalityCode,
                'state': record.get('state'),
                'town': record.get('town'),
                'postCode': record.get('postCode'),
            },
            contactInfo={
                'email': record.get('email'),
                'email2': record.get('email2'),
                'mobile': record.get('mobile'),
                'telephone': record.get('phoneAH'),
            },
            privacyInfo={
                'emailOptOut': record.get('emailOptOut'),
                'marketingOptOut': record.get('marketingOptOut'),
                'smsOptOut': record.get('smsOptOut'),
                'privacyOptIn': record.get('privacyOptIn'),
                'phoneOptOut': record.get('phoneOptOut'),
            },
            documents={
                'passportId': record.get('passportId'),
            },
            metaInfo={
                'notes': record.get('notes'),
                'languageSpokenId': record.get('languageSpokenId'),
                'contracts': record.get('contracts'),
                'accounts': record.get('accounts'),
                'mealCard': record.get('mealCard')
            },
            company={
                'companyId': record.get('companyId')
            },
            registeredProperty=hotelId,
            auraRecordUpdatedOn=datetime.now(),
        )
        return guest

    def _retrieve_data(self, id, name, url):
        result = None
        try:
            logging.debug(f'Extracting {name} for reservation: {id}')
            header = {
                'Content-Type': 'application/json',
                'authtoken': self.token.key
            }
            response = requests.request("GET", url, headers=header)
            if response.status_code == 200:
                result = response.json()
            else:
                logging.error(f'Error getting {name} for reservation {id} due to status code {response.status_code}')
                logging.error(url.format(id))
        except Exception as e:
            logging.error(f' {name} failed due to {e}')
        finally:
            return name, result

    def extract_reservation_details(self, record):
        feature_list = [
            {
                'name': 'holds',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/holds"
            },
            {
                'name': 'guests',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/guests"
            },
            {
                'name': 'billTo',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/billTo"
            },
            {
                'name': 'transfers',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/transfers"
            },
            {
                'name': 'auditTrail',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/auditTrail"
            },
            {
                'name': 'daily_rates',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/dailyRates"
            },
            {
                'name': 'rego_access',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/regoAccess"
            },
            {
                'name': 'requirement',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/requirements"
            },
            {
                'name': 'housekeeping',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/housekeeping"
            },
            {
                'name': 'daily_revenue',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/dailyRevenue"
            },
            {
                'name': 'add_ons',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/reservationAddOn"
            },
            {
                'name': 'correspondence',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/correspondence"
            },
            {
                'name': 'financial_info_actual',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/actualAccount"
            },
            {
                'name': 'bedConfiguration',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/bedConfiguration"
            },
            {
                'name': 'existingToken',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/existingToken"
            },
            {
                'name': 'projectedAccount',
                'url': f"{self.token.hostName}/reservations/{record.get('id')}/projectedAccount"
            },
            {
                'name': 'reservationTypes',
                'url': f"{self.token.hostName}/reservationTypes/{record.get('resTypeId')}"
            },
            {
                'name': 'marketSegment',
                'url': f"{self.token.hostName}/marketSegments/{record.get('marketSegmentId')}"
            },
            {
                'name': 'subMarketSegment',
                'url': f"{self.token.hostName}/marketSegments/{record.get('marketSegmentId')}/subMarketSegments"
            }
        ]
        try:
            logging.debug(f"Extracting reservation {record.get('id')} additional details.")
            threads = []
            with ThreadPoolExecutor(max_workers=4) as executor:
                for item in feature_list:
                    threads.append(executor.submit(self._retrieve_data, record.get("id"), item.get("name"), item.get("url")))

                for th in as_completed(threads):
                    try:
                        key, value = th.result()
                        if key == 'subMarketSegment':
                            value = next((x for x in value if record['subMarketSegmentId'] == x['id']), None)
                        record[key] = value
                    except Exception as e:
                        logging.error(f'Error retrieving addition reservation details for {record.get("id")}')
                        logging.error(e)
        except Exception as e:
            logging.error(f'Error retrieving addition reservation details for {record.get("id")}')
            logging.error(e)
        finally:
            return record

    def _extract_guest_profile(self, guest):
        guestId = 0
        try:
            guestId = guest.get('id')

            _fields = [
                {
                    'name': 'accounts',
                    'url': f"{self.token.hostName}/guests/{guestId}/accounts"
                },
                {
                    'name': 'contracts',
                    'url': f"{self.token.hostName}/guests/{guestId}/contracts"
                },
                {
                    'name': 'mealCard',
                    'url': f'{self.token.hostName}/guests/{guestId}/mealCard'
                }
            ]

            threads = []
            with ThreadPoolExecutor(max_workers=4) as executor:
                for item in _fields:
                    threads.append(
                        executor.submit(self._retrieve_data, guestId, item.get("name"), item.get("url")))

                for th in as_completed(threads):
                    key, value = th.result()
                    guest[key] = value
        except Exception as e:
            logging.error(logging.error(f'Guest profile additional details not found : {guestId}'))
            logging.error(e)
        finally:
            return guest

    def _get_guest_list(self, record):
        logging.debug(f'Processing guest list of size: {len(record.get("guests"))}')
        _guest_list = list()
        try:
            for guest in record.get('guests'):
                _guest_profile = self._transform_guest_profile(self._extract_guest_profile(guest))
                self.mongoV2.save(_guest_profile)
                _guest_list.append(ReservationGuestShort(
                    guest=str(guest.get('id')),
                    primary=guest.get('id') == record.get("guestId"),
                    arrivalTransport=None,
                    departureTransport=None
                ))
        except Exception as e:
            logging.error('Error getting guest list')
            logging.error(e)
        return _guest_list

    def _get_guest_info(self, record):
        guest_info = ReservationGuestInfo(
            adults=record.get('adults'),
            children=record.get('children'),
            infants=record.get('infants'),
            childBuckets=None,
            preRegistered=None,  # TODO : create logic to see if user already existed or not
            guest_list=self._get_guest_list(record)
        )
        return guest_info

    def _get_transaction_info(self, resvId):
        result = None
        try:
            logging.debug(f'Extracting transactions for reservation: {resvId}')
            header = {
                'Content-Type': 'application/json',
                'authtoken': self.token.key
            }
            url = f"{self.token.hostName}/transactions/search?limit=500&offset=0"
            body = {
                'reservationIds': [
                    resvId
                ]
            }
            response = requests.request("POST", url, headers=header, data=json.dumps(body))
            if response.status_code == 200:
                result = response.json()
            else:
                logging.error(f'Error getting transactions for reservation {resvId} due to status code {response.status_code}')
                logging.error(url.format(resvId))
                logging.error(body)
        except Exception as e:
            logging.error(f' {resvId} failed due to {e}')
            logging.error(e)
        finally:
            return result

    def _get_cancellation_policy(self, policyId):
        result = self.mongo.get_collection(self.mongo.CANCELLATION_POLICIES).find_one({'_id': policyId})
        return result

    def transform_reservation(self, record):
        try:
            logging.debug(f'Transforming reservation {record.get("id")}')
            reservationV2 = ReservationModelV2(
                _id=str(record.get('id')),
                orgId=self.orgId,
                idObj=_create_reservation_id_obj(record),
                agent=record.get('agent'),
                datesAndDuration={
                    'createdDate': record.get('createdDate'),
                    'confirmedDate': record.get('confirmedDate'),
                    'arrivalDate': record.get('arrivalDate'),
                    'departureDate': record.get('departureDate'),
                    'cancelledDate': record.get('cancelledDate'),
                    'expectedTimes': {
                        'arrival': record.get('arrivalDate'),
                        'departure': record.get('departureDate')
                    },
                    'originalTimeSpan': None,
                    'lastModifyDateTime': record.get('modifiedDate'),
                    'createBusinessDate': record.get('createdDate').split(' ')[0],
                    'createDateTime': record.get('createdDate'),
                    'eta': record.get('eta'),
                    'event': {
                        'eventName': record.get('eventName'),
                        'eventFinish': record.get('eventFinish'),
                        'eventStart': record.get('eventStart'),
                        'attendees': record.get('attendees'),
                    }
                },
                status=_get_status(record.get('status')),
                correspondence=record.get('correspondence'),
                metaInfo={
                    'allowAutoCheckin': None,
                    'allowMobileCheckout': True,
                    'allowMobileViewFolio': True,
                    'allowPreRegistration': True,
                    'allowedActions': True,
                    'computedReservationStatus': None,
                    'creatorId': record.get('createdById'),
                    'postStayChargeAllowed': False,
                    'preStayChargeAllowed': False,
                    'autoCheckInAllowed': False,
                    'postToNoShowCancelAllowed': False,
                    'stampDutyExists': None,
                    'roomAndTaxPosted': None,
                    'hasOpenFolio': None,
                    'lastModifierId': record.get('modifiedById'),
                    'optedForCommunication': True,
                    'walkIn': False,
                    'printRate': None,
                    'remoteCheckInAllowed': True,
                    'roomNumberLocked': True,
                    'holds': record.get('holds')
                },
                property={
                    'hotelId': str(record.get('clientId')),
                    'hotelName': record.get('propertyName'),
                    'roomId': record.get('areaId'),
                    'roomName': record.get('areaName'),
                    'categoryName': record.get('categoryName'),
                    'numberOfRooms': 1,
                    'bedConfiguration': record.get('bedConfiguration')
                },
                stayInfo={
                    'longTerm': record.get('longTerm'),
                    'businessSegmentId': record.get('businessSegmentId'),
                },
                reservationTypes={
                    'id': record.get('reservationTypes').get('id') if record.get('reservationTypes') else None,
                    'name': record.get('reservationTypes').get('name') if record.get('reservationTypes') else None,
                    'type': record.get('reservationType'),
                    'inactive': record.get('reservationTypes').get('inactive') if record.get('reservationTypes') else None,
                    'fixedRes': record.get('fixedRes'),
                    'fixedResReasonId': record.get('fixedResReasonId'),
                },
                bookingInfo={
                    'upgradeEligible': False,  # TODO : write logic to check if booking is upgrade eligible
                    'upgradeReason': record.get('financial_info_actual').get('upgradeReason'),
                    # 'bookingMedium': record.get('roomStay'),
                    # 'bookingMediumDescription': record.get('roomStay'),
                    'sourceOfSaleType': record.get('bookingSourceName'),
                    'sourceOfSaleCode': record.get('bookingSourceId'),
                    'travelAgentId': record.get('travelAgentId'),
                    'travelAgentName': record.get('travelAgentName'),
                    'marketSegment': record.get('marketSegment'),
                    'subMarketSegment': record.get('subMarketSegment'),
                    'notes': record.get('notes'),
                    'otaNotes': record.get('otaNotes'),
                    'mealPlanId': record.get('mealPlanId'),
                    'otaRef1': record.get('otaRef1'),
                    'otaRef2': record.get('otaRef2'),
                    'otaRef3': record.get('otaRef3')
                },
                financeInfo={
                    'accountId': record.get('accountId'),
                    'rateTypeId': record.get('rateTypeId'),
                    'rateTypeName': record.get('rateTypeName'),
                    'rateOnGroup': record.get('rateOnGroup'),
                    'voucherId': record.get('voucherId'),
                    'billCategoryType': record.get('financial_info_actual').get('billCategoryType'),
                    'createTotalRate': record.get('financial_info_actual').get('createTotalRate'),
                    'preAuthCode': record.get('preAuthCode'),
                    'preAuthAmount': record.get('preAuthAmount'),
                    'preAuthExpDate': record.get('preAuthExpDate'),
                    'totalPoints': None,
                    'totalSpending': {
                        'amountBeforeTax': record.get('financial_info_actual').get('baseRate'),
                        'taxAmount': record.get('financial_info_actual').get('tax'),
                        'totalRate': record.get('financial_info_actual').get('totalRate')
                    },
                    'projectedTotalSpending': {
                        'amountBeforeTax': record.get('projectedAccount').get('balance'),
                        'taxAmount': record.get('projectedAccount').get('tax'),
                        'totalRate': record.get('projectedAccount').get('total')
                    },
                    'paymentMethod': {
                        'paymentModeId': record.get('paymentModeId')
                    },
                    'other': {
                        'arBalance': record.get('financial_info_actual').get('arBalance'),
                        'accommodationBalance': record.get('financial_info_actual').get('accommodationBalance'),
                        'electricityBalance': record.get('financial_info_actual').get('electricityBalance'),
                        'extrasBalance': record.get('financial_info_actual').get('extrasBalance'),
                        'gasBalance': record.get('financial_info_actual').get('gasBalance'),
                        'waterBalance': record.get('financial_info_actual').get('waterBalance'),
                        'phoneBalance': record.get('financial_info_actual').get('phoneBalance'),
                        'internetBalance': record.get('financial_info_actual').get('internetBalance'),
                    },
                    'revenueBucketsInfo': None,
                    'transactions': self._get_transaction_info(record.get('id')),
                    'revenue': {
                            'totalFixedCharge': {
                                'amount': record.get('financial_info_actual').get('baseRate')
                            },
                            'totalPayment': {
                                'amount': record.get('financial_info_actual').get('totalRate')
                            },
                            'roomRevenue': {
                                'amount': sum([item['accommodation'] for item in record.get('daily_revenue')])
                            },
                            'foodAndBevRevenue': {
                                'amount': sum([item['foodAndBeverage'] for item in record.get('daily_revenue')])
                            },
                            'otherRevenue': {
                                'amount': sum([item['other'] for item in record.get('daily_revenue')]),
                            },
                            'totalRevenue': {
                                'amount': sum([item['accommodation'] for item in record.get('daily_revenue')]) + sum(
                                    [item['foodAndBeverage'] for item in record.get('daily_revenue')]) + sum(
                                    [item['other'] for item in record.get('daily_revenue')])
                            },
                            'balance': {
                                'amount': record.get('financial_info_actual').get('accommodationBalance')
                            }
                        },
                    'deposit': {
                        'deposit': record.get('financial_info_actual').get('deposit'),
                        'depositRequiredByDate': record.get('financial_info_actual').get('depositRequiredByDate'),
                        'secondDeposit': record.get('financial_info_actual').get('secondDeposit'),
                        'secondDepositRequiredByDate': record.get('financial_info_actual').get('secondDepositRequiredByDate'),
                    },
                    'discount': [
                        {
                            'id': record.get('financial_info_actual').get('discountId'),
                            'name': record.get('financial_info_actual').get('discountName'),
                            'reason': record.get('financial_info_actual').get('discountReason'),
                            'amount': record.get('financial_info_actual').get('discountAmount')
                        }
                    ],
                    'package': [
                        {
                            'amount': record.get('financial_info_actual').get('package')
                        }
                    ],
                    'commission': {
                        'travelAgentCommissionPercentage': record.get('financial_info_actual').get('travelAgentCommissionPercentage')
                    },
                    # 'billTo': record.get('billTo'),
                    # 'billing': record.get('billing')
                },
                daily_activity=_generate_day_to_day_entry(record),
                linkedReservation=record.get('linkedReservation'),
                guestInfo=self._get_guest_info(record),
                eCertificates=[record.get('existingToken') if record.get('existingToken') else {}],
                historyEvents=record.get('auditTrail'),
                cancellation=[
                    {
                        'reason': None,
                        'cxlDate': record.get('cancelledDate'),
                        'userId': record.get('cancelledById'),
                        'policy': self._get_cancellation_policy(record.get('cancellationPolicyId')),
                        'externalCancelId': record.get('externalCancelId'),
                        'businessLostId': record.get('businessLostId')
                    }
                ],
                housekeeping=record.get('housekeeping'),
                comments=record.get('comments'),
                policies=record.get('reservationPolicies'),
                inventoryItems=record.get('inventoryItems'),
                preferences=[],
                requirement=[],
                rego_access=record.get('rego_access'),
                memberships=None,
                packages=record.get('add_ons'),
                transfers=record.get('transfers'),
                auraRecordUpdatedOn=datetime.now()
            )
            return reservationV2
        except Exception as e:
            logging.error(f"Exception while transforming reservation {record.get('id')}")
            logging.error(e)

    def extract_reservation(self, record):
        reservationId = record.get("reservation")
        logging.info(f'Extracting reservation: {reservationId}')

        header = {
            'Content-Type': 'application/json',
            'authtoken': self.token.key
        }
        url = f"{self.token.hostName}/reservations/{reservationId}?modelType=full"
        response = requests.request("GET", url, headers=header)
        if response.status_code == 200:
            try:
                response_json = dict(response.json())
                response_json['orgId'] = self.orgId
                response_json['agent'] = record.get('agent')
                property_info = self.mongo.get_collection(self.mongo.PROPERTY_COLLECTION).find_one(
                    {"rooms.roomId": str(response_json.get('areaId'))},
                    {'primaryDetails.propertyId': 1, 'name': 1,
                     'hotelId': 1}
                )
                response_json['clientId'] = property_info.get('hotelId') if property_info else 0
                response_json['propertyId'] = property_info.get('primaryDetails').get('propertyId') if property_info else 0
                response_json['propertyName'] = property_info.get('name') if property_info else 'Unknown'
                return response_json
            except Exception as err:
                logging.error(f'{record} load failed due to {err}')
        else:
            logging.error(f"RMS returned status code {response.status_code} "
                          f"for reservation {reservationId} ")
            raise Exception('Reservation not found / unable to retrieve')

    def process(self, reservationId):
        record = {
            'orgId': self.orgId,
            'agent': NAME,
            'reservation': reservationId
        }

        if not validate_key(self.token.validity):
            logging.debug('Refreshing RMS token as it is about to expire')
            self.token = get_key(self.orgId)

        record = self.extract_reservation(record)
        if record:
            record = self.transform_reservation(self.extract_reservation_details(record))
            if record:
                self.mongoV2.save(record)
                store_reservation_to_alvie(record)
            else:
                logging.error(f'Error saving the record {record} as it is None')

    def __del__(self):
        if self.mongo:
            self.mongo.close_connection()



from typing import List

from pydantic import BaseModel


class ResponseList(BaseModel):
    status: bool = False
    message: str = 'default error msg'
    data: List[dict] = []


class ResponseDict(BaseModel):
    status: bool = False
    message: str = 'default error msg'
    data: dict



import json

from upswingutil.pms.oracle import NAME, get_key, validate_key
from upswingutil.db import Mongodb, Firestore, MongodbV2
from upswingutil.schema import Token
from datetime import datetime, timedelta
import requests
import logging
from upswingutil.db.model import ReservationModel, Status, ReservationGuestInfo, ReservationGuestShort, GuestsModel


def _get_status(status):
    if status == 'NoShow':
        return Status.NO_SHOW
    elif status == 'CheckedOut':
        return Status.DEPARTED
    elif status == 'Reserved':
        return Status.RESERVED
    elif status == 'Cancelled':
        return Status.CANCELLED
    elif status == 'InHouse':
        return Status.IN_HOUSE
    else:
        print(status)
        return Status.UNCONFIRMED


def _get_id(record):
    _id = next((x['id'] for x in record.get('reservationIdList') if x['type'] == 'Reservation'), None)
    return _id


def _get_transaction_info(transactions):
    return transactions


def _get_package_info(package):
    del package['transactionDate']
    return package


def _generate_day_to_day_entry(record):
    entry_list = list()
    room_rates = record.get('roomStay').get('roomRates')
    start_date = datetime.fromisoformat(record.get('roomStay').get('arrivalDate'))
    end_date = datetime.fromisoformat(record.get('roomStay').get('departureDate'))
    day_delta = timedelta(days=1)
    for i in range((end_date - start_date).days):
        _rates = next((x for x in room_rates if x['start'] == str((start_date + i * day_delta).date())), None)
        _activity = {
            'date': str((start_date + i * day_delta).date()),
            'day_of_stay': i + 1
        }
        if _rates:
            _activity.update(_rates)
        else:
            print('No entry on given date', ((start_date + i * day_delta).date()))
        entry_list.append(_activity)
    return entry_list


def _get_folio_information(record):
    data = {
        'folioWindows': record.get('folioWindows'),
        'folioHistory': record.get('folioHistory')
    }
    return data


def _get_geo_region(nationalID: int):
    result = 'unknown'
    try:
        mongo = Mongodb('upswing')
        region = mongo.get_collection('countries').find_one({"_id": nationalID}, {"region": 1})
        mongo.close_connection()
        result = region if region else 'unknown'
    except Exception as e:
        logging.error(f'Error getting geo region of nation {nationalID}')
        logging.error(e)
    finally:
        return result


@DeprecationWarning
class ReservationSync:

    def __init__(self, orgId: str, g_cloud_token=None):
        self.orgId = orgId
        self.mongo = Mongodb(orgId)
        self.mongoV2 = MongodbV2(orgId)
        self.token: Token = get_key(self.orgId)
        self.g_cloud_token = g_cloud_token
        self._api_call_counter = 0

    def _extract_guest_profile(self, hotelId, guest):
        try:
            guestId = 0
            for item in guest.get('profileInfo').get('profileIdList'):
                if item.get('type') == 'Profile':
                    guestId = item.get('id')

            if not validate_key(self.token.validity):
                logging.info(f'Refreshing {NAME} token as it is about to expire')
                self.token = get_key(self.orgId, self.token.refreshKey)

            _url = f'{self.token.hostName}/crm/v1/profiles/{guestId}?fetchInstructions=Address' \
                   f'&fetchInstructions=Comment&fetchInstructions=Communication&fetchInstructions=Correspondence' \
                   f'&fetchInstructions=DeliveryMethods&fetchInstructions=FutureReservation' \
                   f'&fetchInstructions=GdsNegotiatedRate&fetchInstructions=HistoryReservation' \
                   f'&fetchInstructions=Indicators&fetchInstructions=Keyword&fetchInstructions=Membership' \
                   f'&fetchInstructions=NegotiatedRate&fetchInstructions=Preference&fetchInstructions=Profile' \
                   f'&fetchInstructions=Relationship&fetchInstructions=SalesInfo&fetchInstructions=Subscriptions' \
                   f'&fetchInstructions=WebUserAccount'
            _profile_data = self.retrieve_data(hotelId, _url)
            if _profile_data:
                idObj = dict()
                for item in _profile_data['profileIdList']:
                    idObj[item.get('type').lower()] = item['id']
                _profile_data['idObj'] = idObj
                _profile_data['birthDate'] = guest.get('birthDate')
                return _profile_data
            else:
                logging.error(f'Guest profile not found : {guestId}')

        except Exception as e:
            logging.error(f'Error while transforming guest {guest}')
            logging.error(e)

    def _get_guest_list(self, hotelId, record):
        logging.info(f'Processing guest list of size: {len(record.get("reservationGuests"))}')
        _guest_list = list()
        for item in record.get('reservationGuests'):
            _guest_profile = self._transform_guest_profile(self._extract_guest_profile(hotelId, item))
            self.mongoV2.save(_guest_profile)
            _guest = ReservationGuestShort(
                guest=_guest_profile,
                primary=item.get('primary'),
                birthDate=item.get('birthDate'),
                arrivalTransport=item.get('arrivalTransport'),
                departureTransport=item.get('departureTransport')
            )
            _guest_list.append(_guest)
        return _guest_list

    def _get_guest_info(self, record):
        guest_info = ReservationGuestInfo(
            adults=record.get('roomStay').get('guestCounts').get('adults'),
            children=record.get('roomStay').get('guestCounts').get('children'),
            infants=0,
            childBuckets=record.get('roomStay').get('guestCounts').get('childBuckets'),
            preRegistered=record.get('preRegistered'),
            guest_list=self._get_guest_list(record.get('hotelId'), record)
        )
        return guest_info

    def _transform_guest_profile(self, record):
        details = record.get('profileDetails')
        guest = GuestsModel(
            _id=record.get('idObj').get('profile'),
            idObj=record.get('idObj'),
            birthDate=record.get('birthDate'),
            profileType=details.get('profileType'),
            statusCode=details.get('statusCode'),
            registeredProperty=details.get('registeredProperty'),
            createDateTime=details.get('createDateTime'),
            creatorId=details.get('creatorId'),
            lastModifyDateTime=details.get('lastModifyDateTime'),
            lastModifierId=details.get('lastModifierId'),
            markForHistory=details.get('markForHistory'),
            keywords=details.get('keywords').get('keyword'),
            emails=details.get('emails'),
            telephones=details.get('telephones'),
            addresses=details.get('addresses'),
            profileMemberships=details.get('profileMemberships'),
            relationships=details.get('relationships'),
            stayReservationInfoList=details.get('stayReservationInfoList'),
            lastStayInfo=details.get('lastStayInfo'),
            profileAccessType=details.get('profileAccessType'),
            profileRestrictions=details.get('profileRestrictions'),
            privacyInfo=details.get('privacyInfo'),
            taxInfo=details.get('taxInfo'),
            salesInfo=details.get('salesInfo'),
            subscriptions=details.get('subscriptions'),
            mailingActions=details.get('mailingActions'),
            relationshipsSummary=details.get('relationshipsSummary'),
            preferenceCollection=details.get('preferenceCollection'),
            comments=details.get('comments'),
            profileDeliveryMethods=details.get('profileDeliveryMethods'),
            profileIndicators=details.get('profileIndicators'),
            customer=details.get('customer'),
            company=details.get('company'),
        )

        if details.get('profileType') == 'Guest':
            guest.firstName = details.get('customer').get('personName')[0].get('givenName')
            guest.middleName = details.get('customer').get('personName')[0].get('nameSuffix')
            guest.lastName = details.get('customer').get('personName')[0].get('surname')
        elif details.get('profileType') == 'Group':
            guest.firstName = details.get('company').get('companyName')
        else:
            logging.error(
                f'Unhandled profile type : {details.get("profileType")} for guest {record.get("idObj").get("profile")}')
        return guest

    def _get_reservation_id(self, idList, agent, hotelId, resv):
        resv_dict = dict()
        resv_dict['globalId'] = f'{agent}-{self.orgId}-{hotelId}-{resv}'
        for item in idList:
            value = item['id']
            _type = item['type']
            resv_dict[_type.lower()] = value
        return resv_dict

    def retrieve_data(self, clientId, url, payload="") -> dict:
        headers = {
            'Content-Type': 'application/json',
            'x-hotelid': clientId,
            'x-app-key': self.token.appKey,
            'Authorization': f'Bearer {self.token.key}'
        }
        response = requests.request("GET", url, headers=headers, data=payload)
        self._api_call_counter += 1
        if response.status_code == 200:
            return dict(response.json())
        else:
            logging.error(f'Failed to retrieve data, status: {response.status_code}')
            logging.error(response.reason)
            return None

    def _extract_reservation_details(self, record):
        resv_id = record.get('id')
        try:
            hotel_id = record.get('hotelId')
            record['idObj'] = self._get_reservation_id(record.get('reservationIdList'), record.get('agent'), record.get('hotelId'), record.get('id'))
            logging.info(f"Extracting reservation {resv_id} additional details.")
            if self.token is None:
                self.token: Token = record.get('token')

            if not validate_key(self.token.validity):
                logging.info(f'Refreshing {NAME} token as it is about to expire')
                self.token = get_key(self.orgId, self.token.refreshKey)

            _url = f'{self.token.hostName}/rsv/v0/hotels/{hotel_id}/reservations/{resv_id}/cancellationHistory'
            data = self.retrieve_data(record.get('hotelId'), _url)
            record['cancellation'] = data.get('cxlActivityLog') if data else None

            _url = f'{self.token.hostName}/rsv/v0/hotels/{hotel_id}/reservations/{resv_id}/linkedSummary'
            data = self.retrieve_data(record.get('hotelId'), _url)
            record['linkedReservation'] = data.get('linkedReservationList') if data else None

            _url = f'{self.token.hostName}/rsv/v0/hotels/{hotel_id}/reservations/{resv_id}/calls'
            data = self.retrieve_data(record.get('hotelId'), _url)
            record['callHistory'] = data.get('callHistory') if data else None

            _url = f'{self.token.hostName}/rsv/v0/hotels/{hotel_id}/reservations/{resv_id}/inventoryItems'
            data = self.retrieve_data(record.get('hotelId'), _url)
            record['inventoryItems'] = data.get('inventoryItems') if data else None

            _url = f'{self.token.hostName}/csh/v0/hotels/{hotel_id}/reservations/{resv_id}/folios?includeFolioHistory=true&fetchInstructions=Account&fetchInstructions=Payee&fetchInstructions=Totalbalance&fetchInstructions=Windowbalances&fetchInstructions=Payment&fetchInstructions=Postings&fetchInstructions=Transactioncodes&fetchInstructions=Reservation'
            data = self.retrieve_data(record.get('hotelId'), _url)
            record['folioInformation'] = data.get('reservationFolioInformation') if data else None

            _url = f'{self.token.hostName}/csh/v0/hotels/{hotel_id}/transactions?reservationList={resv_id}&idContext=OPERA&type=Reservation&includeGenerates=true&includeTransactionsWithFolioNo=true&includeTransactionsWithManualPostingOnly=true'
            data = self.retrieve_data(record.get('hotelId'), _url)
            if data:
                del data['links']
            record['transactions'] = data

            return record
        except Exception as e:
            logging.error(f'Error retrieving addition reservation details for {resv_id}')
            logging.error(e)

    def _extract_reservation(self, record):
        try:
            reservationId = record.get("reservation")["reservation"]
            hotelId = record.get('reservation')['hotelId']
            logging.info(f'Extracting reservation: {reservationId}')

            header = {
                'Content-Type': 'application/json',
                'x-hotelid': hotelId,
                'x-app-key': self.token.appKey,
                'Authorization': f'Bearer {self.token.key}'
            }

            url = f"{self.token.hostName}/rsv/v1/hotels/{hotelId}/reservations/{reservationId}?fetchInstructions" \
                  f"=Comments&fetchInstructions=GuestMemberships&fetchInstructions=GuestLastStay&fetchInstructions" \
                  f"=ProfileAwards&fetchInstructions=ScheduledActivities&fetchInstructions=ServiceRequests" \
                  f"&fetchInstructions=ReservationAwards&fetchInstructions=RevenuesAndBalances&fetchInstructions" \
                  f"=Tickets&fetchInstructions=GuestComments&fetchInstructions=Packages&fetchInstructions" \
                  f"=InventoryItems&fetchInstructions=ReservationPaymentMethods&fetchInstructions=RoutingInstructions" \
                  f"&fetchInstructions=Preferences&fetchInstructions=Memberships&fetchInstructions=Alerts" \
                  f"&fetchInstructions=Traces&fetchInstructions=ConfirmationLetters&fetchInstructions=CallHistory" \
                  f"&fetchInstructions=FixedCharges&fetchInstructions=GuestMessages&fetchInstructions" \
                  f"=ReservationPolicies&fetchInstructions=Indicators&fetchInstructions=LinkedReservations" \
                  f"&fetchInstructions=ECoupons&fetchInstructions=TrackItItems&fetchInstructions=WebRegistrationCards" \
                  f"&fetchInstructions=ServiceRequests&fetchInstructions=ReservationActivities&fetchInstructions" \
                  f"=PrepaidCards&fetchInstructions=Shares&fetchInstructions=Attachments&fetchInstructions=Locators" \
                  f"&fetchInstructions=TransactionDiversions&fetchInstructions=ECertificates&fetchInstructions" \
                  f"=UpsellInfo&fetchInstructions=RoomAssignedByAI&fetchInstructions=Reservation"
            response = requests.request("GET", url, headers=header)
            self._api_call_counter += 1
            if response.status_code == 200:
                response_json = dict(response.json()).get('reservations').get('reservation')
                if len(response_json) > 0:
                    result = response_json[0]
                    result['id'] = reservationId
                    result['token'] = self.token
                    result['orgId'] = record.get('orgId')
                    result['agent'] = record.get('agent')
                    return result
                else:
                    logging.error(f"{NAME} returned reservation {reservationId} of length zero")
            else:
                logging.error(f"{NAME} returned status code {response.status_code} for reservation {reservationId} ")
        except Exception as err:
            logging.error(f'{record} load failed due to {err}')
        return None

    def _transform_reservation(self, record):
        reservationV2 = ReservationModel(
            _id=record.get('id'),
            hotelId=record.get('hotelId'),
            orgId=self.orgId,
            idObj=record.get('idObj'),
            agent=record.get('agent'),
            hotelName=record.get('folioInformation').get('reservationInfo').get('hotelName'),
            arrivalDate=record.get('roomStay').get('arrivalDate'),
            departureDate=record.get('roomStay').get('departureDate'),
            expectedTimes={
                'arrival': record.get('roomStay').get('expectedTimes').get('reservationExpectedArrivalTime'),
                'departure': record.get('roomStay').get('expectedTimes').get('reservationExpectedDepartureTime')
            },
            originalTimeSpan=record.get('roomStay').get('originalTimeSpan'),
            status=_get_status(record.get('reservationStatus')),
            alerts=record.get('alerts'),
            metaInfo={
                'allowAutoCheckin': record.get('allowAutoCheckin'),
                'allowMobileCheckout': record.get('allowMobileCheckout'),
                'allowMobileViewFolio': record.get('allowMobileViewFolio'),
                'allowPreRegistration': record.get('allowPreRegistration'),
                'allowedActions': record.get('allowedActions'),
                'computedReservationStatus': record.get('computedReservationStatus'),
                'creatorId': record.get('creatorId'),
                'postStayChargeAllowed': record.get('folioInformation').get('postStayChargeAllowed'),
                'preStayChargeAllowed': record.get('folioInformation').get('preStayChargeAllowed'),
                'autoCheckInAllowed': record.get('folioInformation').get('autoCheckInAllowed'),
                'postToNoShowCancelAllowed': record.get('folioInformation').get('postToNoShowCancelAllowed'),
                'stampDutyExists': record.get('folioInformation').get('stampDutyExists'),
                'roomAndTaxPosted': record.get('folioInformation').get('roomAndTaxPosted'),
                'hasOpenFolio': record.get('hasOpenFolio'),
                'lastModifierId': record.get('lastModifierId'),
                'optedForCommunication': record.get('optedForCommunication'),
                'walkIn': record.get('walkIn'),
                'printRate': record.get('printRate'),
                'remoteCheckInAllowed': record.get('roomStay').get('remoteCheckInAllowed'),
                'roomNumberLocked': record.get('roomStay').get('roomNumberLocked'),
                'roomStayReservation': record.get('roomStayReservation'),
                'reservationIndicators': record.get('reservationIndicators'),
                'routingInstructions': record.get('routingInstructions')
            },
            createBusinessDate=record.get('createBusinessDate'),
            createDateTime=record.get('createDateTime'),
            guestLocators=record.get('guestLocators'),
            lastModifyDateTime=record.get('lastModifyDateTime'),
            bookingInfo={
                'upgradeEligible': record.get('upgradeEligible'),
                'bookingMedium': record.get('roomStay').get('bookingMedium'),
                'bookingMediumDescription': record.get('roomStay').get('bookingMediumDescription'),
                'guarantee': record.get('roomStay').get('guarantee'),
                'sourceOfSaleType': record.get('sourceOfSale').get('sourceType'),
                'sourceOfSaleCode': record.get('sourceOfSale').get('sourceCode'),
            },
            financeInfo={
                'totalPoints': record.get('roomStay').get('totalPoints'),
                'totalSpending': record.get('roomStay').get('total'),
                'paymentMethod': record.get('reservationPaymentMethods'),
                'revenueBucketsInfo': record.get('revenueBucketsInfo'),
                'transactions': _get_transaction_info(record.get('transactions')),
                'revenue': record.get('cashiering')
            },
            daily_activity=_generate_day_to_day_entry(record),
            linkedReservation=record.get('linkedReservation'),
            guestInfo=self._get_guest_info(record),
            eCertificates=record.get('eCertificates'),
            historyEvents=record.get('callHistory'),
            cancellation=record.get('cancellation'),
            comments=record.get('comments'),
            policies=record.get('reservationPolicies'),
            inventoryItems=record.get('inventoryItems'),
            preferences=record.get('preferenceCollection'),
            memberships=record.get('reservationMemberships'),
            packages=record.get('reservationPackages'),
            auraRecordUpdatedOn=str(datetime.now())
        )

        if record.get('folioInformation'):
            _f_info = record.get('folioInformation')
            reservationV2.folioInformation = _get_folio_information(_f_info)
            reservationV2.roomStay = _f_info.get('reservationInfo').get('roomStay')
            reservationV2.financeInfo['taxType'] = _f_info.get('reservationInfo').get('taxType')
            reservationV2.financeInfo['commissionPayoutTo'] = _f_info.get('reservationInfo').get('commissionPayoutTo'),
            del _f_info

        return reservationV2

    def _load_to_mongodb(self, record):

        try:
            logging.info(f'Adding reservation {record.id} to db {record.orgId}')
            self.mongo.get_collection(self.mongo.RESERVATION_COLLECTION).update_one({'_id': record.id},
                                                                          {'$set': record.dict(exclude={'token', 'id'})}, upsert=True)
        except Exception as e:
            logging.error(f'Error storing to respective {record.id} to db {record.orgId}')
            logging.error(e)

    def _load_to_alvie(self, record: ReservationModel):
        firestore_db = Firestore(app='alvie')
        guest_list = [item["idObj"] for item in record.guestInfo.get('guest_list')]
        user_email_list = []
        resv_info = {
            "_id": record.id,
            "agent": record.agent,
            "orgId": record.orgId,
            "areaId": record.roomStay,
            "arrivalDate": record.arrivalDate,
            "departureDate": record.departureDate,
            "propertyId": record.hotelId,
            "propertyName": record.hotelName,
            "status": record.status,
            "travelAgentId": record.bookingInfo.get('sourceOfSaleCode'),
            "travelAgentName": record.bookingInfo.get('sourceOfSaleType')
        }
        logging.debug(f"Final reservation id : {record.id}")
        for email in user_email_list:
            docs = firestore_db.get_collection('users').where("email", "==", email).stream()
            for doc in docs:
                logging.info("loading reservation to alvie")
                firestore_db.get_collection(f'users/{doc.id}/reservations') \
                    .document(str(record.id)) \
                    .set(resv_info, merge=True)
                logging.debug(f"added to {doc.id}")

    def process(self, reservationId, hotelId):
        record = {
            'orgId': self.orgId,
            'agent': NAME,
            'reservation': {
                'reservation': reservationId,
                'hotelId': hotelId
            }
        }

        if not validate_key(self.token.validity):
            logging.info(f'Refreshing {NAME} token as it is about to expire')
            self.token = get_key(self.orgId, self.token.refreshKey)

        record = self._extract_reservation(record)
        if record:
            record = self._extract_reservation_details(record)
            record = self._transform_reservation(record)
            self.mongoV2.save(record)
            # self._load_to_alvie(record)

    def __del__(self):
        logging.info(f'API calls made: {self._api_call_counter}')
        if self.mongo:
            self.mongo.close_connection()
        if self.mongoV2:
            self.mongoV2.close_connection()



COUNTRIES = [
    {
        "id": 1,
        "name": "Afghanistan"
    },
    {
        "id": 2,
        "name": "Algeria"
    },
    {
        "id": 3,
        "name": "American Samoa"
    },
    {
        "id": 4,
        "name": "Andorra"
    },
    {
        "id": 5,
        "name": "Angola"
    },
    {
        "id": 7,
        "name": "Albania"
    },
    {
        "id": 8,
        "name": "Antarctica"
    },
    {
        "id": 10,
        "name": "Argentina"
    },
    {
        "id": 11,
        "name": "Armenia"
    },
    {
        "id": 12,
        "name": "Aruba"
    },
    {
        "id": 13,
        "name": "Australia"
    },
    {
        "id": 14,
        "name": "Austria"
    },
    {
        "id": 15,
        "name": "Azerbaidjan"
    },
    {
        "id": 17,
        "name": "Bangladesh"
    },
    {
        "id": 19,
        "name": "Belgium"
    },
    {
        "id": 20,
        "name": "Belize"
    },
    {
        "id": 21,
        "name": "Benin"
    },
    {
        "id": 24,
        "name": "Bolivia"
    },
    {
        "id": 25,
        "name": "Bosnia Herzegovina"
    },
    {
        "id": 26,
        "name": "Botswana"
    },
    {
        "id": 28,
        "name": "Brazil"
    },
    {
        "id": 30,
        "name": "Brunei Darussalam"
    },
    {
        "id": 31,
        "name": "Bulgaria"
    },
    {
        "id": 32,
        "name": "Bahrain"
    },
    {
        "id": 33,
        "name": "Burkina Faso"
    },
    {
        "id": 34,
        "name": "Burundi"
    },
    {
        "id": 35,
        "name": "Buthan"
    },
    {
        "id": 36,
        "name": "Cambodia"
    },
    {
        "id": 37,
        "name": "Cameroon"
    },
    {
        "id": 38,
        "name": "Canada"
    },
    {
        "id": 39,
        "name": "Cape Verde"
    },
    {
        "id": 41,
        "name": "Central African Rep"
    },
    {
        "id": 42,
        "name": "Chad"
    },
    {
        "id": 43,
        "name": "Chile"
    },
    {
        "id": 44,
        "name": "China"
    },
    {
        "id": 47,
        "name": "Colombia"
    },
    {
        "id": 48,
        "name": "Congo"
    },
    {
        "id": 49,
        "name": "Cook Islands"
    },
    {
        "id": 50,
        "name": "Costa Rica"
    },
    {
        "id": 51,
        "name": "Cuba"
    },
    {
        "id": 52,
        "name": "Cyprus"
    },
    {
        "id": 53,
        "name": "Czech Republic"
    },
    {
        "id": 54,
        "name": "Djibouti"
    },
    {
        "id": 55,
        "name": "Denmark"
    },
    {
        "id": 57,
        "name": "Ecuador"
    },
    {
        "id": 58,
        "name": "Egypt"
    },
    {
        "id": 59,
        "name": "El Salvador"
    },
    {
        "id": 60,
        "name": "Equatorial Guinea"
    },
    {
        "id": 61,
        "name": "Estonia"
    },
    {
        "id": 62,
        "name": "Ethiopia"
    },
    {
        "id": 63,
        "name": "Finland"
    },
    {
        "id": 64,
        "name": "Falkland Island"
    },
    {
        "id": 65,
        "name": "Faroe Islands"
    },
    {
        "id": 66,
        "name": "Fiji"
    },
    {
        "id": 67,
        "name": "France"
    },
    {
        "id": 70,
        "name": "Gabon"
    },
    {
        "id": 71,
        "name": "Georgia"
    },
    {
        "id": 72,
        "name": "Germany"
    },
    {
        "id": 73,
        "name": "Ghana"
    },
    {
        "id": 74,
        "name": "Gibraltar"
    },
    {
        "id": 76,
        "name": "Greece"
    },
    {
        "id": 77,
        "name": "Guatemala"
    },
    {
        "id": 78,
        "name": "Greenland"
    },
    {
        "id": 79,
        "name": "Gambia"
    },
    {
        "id": 81,
        "name": "Guadeloupe"
    },
    {
        "id": 83,
        "name": "Guinea"
    },
    {
        "id": 84,
        "name": "Guinea Bissau"
    },
    {
        "id": 85,
        "name": "Guyana"
    },
    {
        "id": 86,
        "name": "Hong Kong"
    },
    {
        "id": 87,
        "name": "Haiti"
    },
    {
        "id": 89,
        "name": "Honduras"
    },
    {
        "id": 90,
        "name": "Croatia"
    },
    {
        "id": 91,
        "name": "Hungary"
    },
    {
        "id": 92,
        "name": "Iceland"
    },
    {
        "id": 93,
        "name": "India"
    },
    {
        "id": 94,
        "name": "Indonesia"
    },
    {
        "id": 95,
        "name": "Iran"
    },
    {
        "id": 96,
        "name": "Iraq"
    },
    {
        "id": 97,
        "name": "Ireland"
    },
    {
        "id": 98,
        "name": "Israel"
    },
    {
        "id": 99,
        "name": "Italy"
    },
    {
        "id": 102,
        "name": "Japan"
    },
    {
        "id": 103,
        "name": "Jordan"
    },
    {
        "id": 104,
        "name": "Kazachstan"
    },
    {
        "id": 105,
        "name": "Kenya"
    },
    {
        "id": 106,
        "name": "Kyrgyzstan"
    },
    {
        "id": 107,
        "name": "Kiribati"
    },
    {
        "id": 108,
        "name": "Comoros"
    },
    {
        "id": 109,
        "name": "Korea North"
    },
    {
        "id": 110,
        "name": "Korea South"
    },
    {
        "id": 111,
        "name": "Kuwait"
    },
    {
        "id": 112,
        "name": "Laos"
    },
    {
        "id": 113,
        "name": "Latvia"
    },
    {
        "id": 114,
        "name": "Lebanon"
    },
    {
        "id": 115,
        "name": "Lesotho"
    },
    {
        "id": 116,
        "name": "Liberia"
    },
    {
        "id": 117,
        "name": "Libya"
    },
    {
        "id": 118,
        "name": "Liechtenstein"
    },
    {
        "id": 119,
        "name": "Lithuania"
    },
    {
        "id": 120,
        "name": "Luxembourg"
    },
    {
        "id": 121,
        "name": "Macau"
    },
    {
        "id": 122,
        "name": "Madagascar"
    },
    {
        "id": 123,
        "name": "Malawi"
    },
    {
        "id": 124,
        "name": "Malaysia"
    },
    {
        "id": 125,
        "name": "Maldives"
    },
    {
        "id": 126,
        "name": "Mali"
    },
    {
        "id": 127,
        "name": "Marshall Islands"
    },
    {
        "id": 128,
        "name": "Martinique"
    },
    {
        "id": 129,
        "name": "Mauritania"
    },
    {
        "id": 130,
        "name": "Mauritius"
    },
    {
        "id": 131,
        "name": "Mexico"
    },
    {
        "id": 132,
        "name": "Micronesia"
    },
    {
        "id": 133,
        "name": "Moldavia"
    },
    {
        "id": 134,
        "name": "Monaco"
    },
    {
        "id": 136,
        "name": "Malta"
    },
    {
        "id": 137,
        "name": "Morocco"
    },
    {
        "id": 138,
        "name": "Mozambique"
    },
    {
        "id": 139,
        "name": "Myanmar"
    },
    {
        "id": 140,
        "name": "Mongolia"
    },
    {
        "id": 141,
        "name": "Namibia"
    },
    {
        "id": 142,
        "name": "Nauru"
    },
    {
        "id": 143,
        "name": "Nepal"
    },
    {
        "id": 144,
        "name": "Netherland Antilles"
    },
    {
        "id": 145,
        "name": "Netherlands"
    },
    {
        "id": 147,
        "name": "New Caledonia"
    },
    {
        "id": 148,
        "name": "Nicaragua"
    },
    {
        "id": 149,
        "name": "Niger"
    },
    {
        "id": 150,
        "name": "Nigeria"
    },
    {
        "id": 151,
        "name": "Niue"
    },
    {
        "id": 152,
        "name": "New Zealand"
    },
    {
        "id": 153,
        "name": "Norfolk Island"
    },
    {
        "id": 155,
        "name": "Norway"
    },
    {
        "id": 156,
        "name": "Oman"
    },
    {
        "id": 157,
        "name": "Pakistan"
    },
    {
        "id": 158,
        "name": "Palau"
    },
    {
        "id": 159,
        "name": "Panama"
    },
    {
        "id": 160,
        "name": "Papua New Guinea"
    },
    {
        "id": 161,
        "name": "Paraguay"
    },
    {
        "id": 162,
        "name": "Peru"
    },
    {
        "id": 163,
        "name": "Philippines"
    },
    {
        "id": 165,
        "name": "Poland"
    },
    {
        "id": 166,
        "name": "French Polynesia"
    },
    {
        "id": 167,
        "name": "Portugal"
    },
    {
        "id": 169,
        "name": "Qatar"
    },
    {
        "id": 170,
        "name": "Reunion"
    },
    {
        "id": 171,
        "name": "Romania"
    },
    {
        "id": 172,
        "name": "Russia"
    },
    {
        "id": 173,
        "name": "Rwanda"
    },
    {
        "id": 176,
        "name": "San Marino"
    },
    {
        "id": 177,
        "name": "Saudi Arabia"
    },
    {
        "id": 178,
        "name": "Senegal"
    },
    {
        "id": 179,
        "name": "Seychelles"
    },
    {
        "id": 180,
        "name": "Sudan"
    },
    {
        "id": 181,
        "name": "Sierra_Leone"
    },
    {
        "id": 182,
        "name": "Slovakia"
    },
    {
        "id": 183,
        "name": "Slovenia"
    },
    {
        "id": 184,
        "name": "Solomon Islands"
    },
    {
        "id": 185,
        "name": "Somalia"
    },
    {
        "id": 186,
        "name": "Suriname"
    },
    {
        "id": 187,
        "name": "South Africa"
    },
    {
        "id": 189,
        "name": "Spain"
    },
    {
        "id": 190,
        "name": "Sri Lanka"
    },
    {
        "id": 191,
        "name": "St Helena"
    },
    {
        "id": 192,
        "name": "St Pierre And Miquelon"
    },
    {
        "id": 193,
        "name": "Sao Tome and Principe"
    },
    {
        "id": 196,
        "name": "Swaziland"
    },
    {
        "id": 197,
        "name": "Sweden"
    },
    {
        "id": 198,
        "name": "Singapore"
    },
    {
        "id": 199,
        "name": "Switzerland"
    },
    {
        "id": 200,
        "name": "Syria"
    },
    {
        "id": 201,
        "name": "Tadjikistan"
    },
    {
        "id": 202,
        "name": "Taiwan"
    },
    {
        "id": 203,
        "name": "Tanzania"
    },
    {
        "id": 204,
        "name": "Ukraine"
    },
    {
        "id": 205,
        "name": "Thailand"
    },
    {
        "id": 206,
        "name": "Togolese Republic"
    },
    {
        "id": 207,
        "name": "Tokelau"
    },
    {
        "id": 208,
        "name": "Tonga Islands"
    },
    {
        "id": 209,
        "name": "East Timor"
    },
    {
        "id": 211,
        "name": "Tunisia"
    },
    {
        "id": 212,
        "name": "Turkey"
    },
    {
        "id": 213,
        "name": "Turkmenistan"
    },
    {
        "id": 215,
        "name": "Tuvalu"
    },
    {
        "id": 216,
        "name": "Uganda"
    },
    {
        "id": 217,
        "name": "United Arab Emirates"
    },
    {
        "id": 218,
        "name": "United Kingdom"
    },
    {
        "id": 219,
        "name": "United States"
    },
    {
        "id": 220,
        "name": "Uruguay"
    },
    {
        "id": 222,
        "name": "Uzbekistan"
    },
    {
        "id": 223,
        "name": "Vanuatu"
    },
    {
        "id": 224,
        "name": "Vatican City State"
    },
    {
        "id": 225,
        "name": "Venezuela"
    },
    {
        "id": 226,
        "name": "Vietnam"
    },
    {
        "id": 229,
        "name": "Wallis And Futuna Island"
    },
    {
        "id": 230,
        "name": "Samoa"
    },
    {
        "id": 231,
        "name": "Yemen"
    },
    {
        "id": 232,
        "name": "Yugoslavia"
    },
    {
        "id": 234,
        "name": "Zambia"
    },
    {
        "id": 235,
        "name": "Zimbabwe"
    },
    {
        "id": 236,
        "name": "Zaire"
    },
    {
        "id": 237,
        "name": "Barbados"
    },
    {
        "id": 238,
        "name": "Jamaica"
    },
    {
        "id": 239,
        "name": "South Sudan"
    },
    {
        "id": 243,
        "name": "Palestine"
    },
    {
        "id": 244,
        "name": "Belarus"
    },
    {
        "id": 245,
        "name": "England"
    },
    {
        "id": 246,
        "name": "Scotland"
    },
    {
        "id": 247,
        "name": "Wales"
    },
    {
        "id": 248,
        "name": "Northern Ireland"
    },
    {
        "id": 249,
        "name": "Dominican Republic"
    },
    {
        "id": 250,
        "name": "Montenegro"
    },
    {
        "id": 251,
        "name": "Republic Of Macedonia"
    },
    {
        "id": 252,
        "name": "Anguilla"
    },
    {
        "id": 253,
        "name": "Antigua And Barbuda"
    },
    {
        "id": 254,
        "name": "Bahamas"
    },
    {
        "id": 255,
        "name": "Bermuda"
    },
    {
        "id": 256,
        "name": "British Indian Ocean Territory"
    },
    {
        "id": 257,
        "name": "British Virgin Islands"
    },
    {
        "id": 258,
        "name": "Cayman Islands"
    },
    {
        "id": 259,
        "name": "Christmas Island"
    },
    {
        "id": 260,
        "name": "Cocos Islands"
    },
    {
        "id": 261,
        "name": "Curacao"
    },
    {
        "id": 262,
        "name": "Dominica"
    },
    {
        "id": 263,
        "name": "Eritrea"
    },
    {
        "id": 264,
        "name": "Grenada"
    },
    {
        "id": 265,
        "name": "Guam"
    },
    {
        "id": 266,
        "name": "Guernsey"
    },
    {
        "id": 267,
        "name": "Isle Of Man"
    },
    {
        "id": 268,
        "name": "Ivory Coast"
    },
    {
        "id": 269,
        "name": "Jersey"
    },
    {
        "id": 270,
        "name": "Kosovo"
    },
    {
        "id": 271,
        "name": "Mayotte"
    },
    {
        "id": 272,
        "name": "Montserrat"
    },
    {
        "id": 273,
        "name": "Northern Mariana Islands"
    },
    {
        "id": 274,
        "name": "Pitcairn"
    },
    {
        "id": 275,
        "name": "Puerto Rico"
    },
    {
        "id": 276,
        "name": "Saint Barthelemy"
    },
    {
        "id": 277,
        "name": "Saint Kitts And Nevis"
    },
    {
        "id": 278,
        "name": "Saint Lucia"
    },
    {
        "id": 279,
        "name": "Saint Martin"
    },
    {
        "id": 280,
        "name": "Saint Vincent And The Grenadines"
    },
    {
        "id": 281,
        "name": "Serbia"
    },
    {
        "id": 282,
        "name": "Sint Maarten"
    },
    {
        "id": 283,
        "name": "Svalbard And JanMayen"
    },
    {
        "id": 284,
        "name": "Trinidad And Tobago"
    },
    {
        "id": 285,
        "name": "Turks And Caicos Islands"
    },
    {
        "id": 286,
        "name": "US Virgin Islands"
    },
    {
        "id": 287,
        "name": "Western Sahara"
    }
]



from fastapi import Depends, HTTPException, requests
from fastapi.security import HTTPBearer
from firebase_admin import auth
from starlette import status

jwtBearer = HTTPBearer()


def verify_and_decode_jwt(token: dict = Depends(jwtBearer)):
    try:
        user = auth.verify_id_token(token.credentials)
        org = user.get('o')
        role = user.get('r')
        requests.org = org
        requests.role = role
    except Exception as e:
        print(e)
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid JWT",
        )


def verify_owner_and_get_id(token: dict = Depends(jwtBearer)):
    try:
        user = auth.verify_id_token(token.credentials)
        org = user.get('o')
        role = user.get('r')
        requests.owner = user.get('email')
        requests.org = org
        requests.role = role
    except Exception as e:
        print(e)
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid JWT",
        )


def verify_and_get_jwt(token: dict = Depends(jwtBearer)):
    try:
        user = auth.verify_id_token(token.credentials)
        org = user.get('o')
        role = user.get('r')
        requests.org = org
        requests.role = role
        requests.token = token
    except Exception as e:
        print(e)
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid JWT",
        )


import json
import logging
import unittest

import firebase_admin
from firebase_admin import credentials

import upswingutil as ul
from upswingutil.integrations.messerschmitt import get_key
from upswingutil.resource import access_secret_version
from upswingutil.schema import Token

ul.G_CLOUD_PROJECT = "aura-staging-31cae"
ul.ENCRYPTION_SECRET = access_secret_version(ul.G_CLOUD_PROJECT, 'ENCRYPT_SECRET_KEY', '1')


try:
    firebase_admin.get_app()
except Exception as e:
    logging.info('Initializing default firebase app')
    cred = json.loads(access_secret_version(ul.G_CLOUD_PROJECT, 'Firebase-Alvie', '1'))
    cred = credentials.Certificate(cred)
    firebase = firebase_admin.initialize_app(cred)


try:
    firebase_admin.get_app('alvie')
except Exception as e:
    logging.info('Initializing default firebase app')
    cred = json.loads(access_secret_version(ul.G_CLOUD_PROJECT, 'Firebase-Alvie', '1'))
    cred = credentials.Certificate(cred)
    firebase = firebase_admin.initialize_app(name='alvie', credential=cred)


class TestMesserschmitt(unittest.TestCase):

    def test_get_key(self):
        orgId = '11281'
        token: Token = get_key(orgId)
        print(token.json())

    def test_get_key_with_app(self):
        orgId = '11281'
        token: Token = get_key(orgId, 'alvie')
        print(token.json())




from google.cloud import secretmanager


def add_secret_version(project_id, secret_id, payload):
    """
    Add a new secret version to the given secret with the provided payload.
    """

    # Create the Secret Manager client.
    client = secretmanager.SecretManagerServiceClient()

    # Build the resource name of the parent secret.
    parent = client.secret_path(project_id, secret_id)

    # Convert the string payload into a bytes. This step can be omitted if you
    # pass in bytes instead of a str for the payload argument.
    payload = payload.encode("UTF-8")

    # Add the secret version.
    response = client.add_secret_version(
        request={"parent": parent, "payload": {"data": payload}}
    )

    # Print the new secret version name.
    print("Added secret version: {}".format(response.name))


def access_secret_version(project_id, secret_id, version_id):
    """
    Access the payload for the given secret version if one exists. The version
    can be a version number as a string (e.g. "5") or an alias (e.g. "latest").
    """

    # Create the Secret Manager client.
    client = secretmanager.SecretManagerServiceClient()

    # Build the resource name of the secret version.
    name = f"projects/{project_id}/secrets/{secret_id}/versions/{version_id}"

    # Access the secret version.
    response = client.access_secret_version(request={"name": name})

    payload = response.payload.data.decode("UTF-8")
    return payload



import ssl
import os
from pymongo import MongoClient
from loguru import logger
import upswingutil as ul

__PII_FIELDS__ = ['guestGiven', 'guestSurname', 'mobile', 'passportId', 'email', 'email2']


class Mongodb:
    RESOURCE_COLLECTION = 'resource'
    AREAS_COLLECTION = "areas"
    GUEST_COLLECTION = "guests"
    PROPERTY_COLLECTION = "properties"
    RESERVATION_COLLECTION = "reservations"
    TRANSACTION_COLLECTION = "transactions"
    AREA_REPORTS_BY_DAY_COLLECTION = "area_report_by_day"
    COUNTRIES_COLLECTION = 'countries'
    ORG_CONFIG_COLLECTION = 'org_config'
    COUNTERS_COLLECTION = 'counters'
    EVENTS_COLLECTION = 'events'
    CANCELLATION_POLICIES = 'cancellationPolicies'
    DAILY_REPORTS_COLLECTION = 'daily_report'
    OFFERS_COLLECTION = 'offers'
    OFFERS_CAMPAIGN_COLLECTION = 'offers_campaigns'
    OFFERS_FEEDBACK_COLLECTION = 'offers_feedback'
    OTA_CHANNEL_COLLECTION = 'ota_channel'
    IOT_ROOM_MAPPING = 'iot_room_mapping'
    INTEGRATION_PROPERTY = 'integration_property'
    INTEGRATION_ORG = 'integration_org'
    OFFERS_TARGETGROUP_COLLECTION = 'offers_targetgroups'
    UPSWING_INTEGRATIONS = 'integrations'


    def __init__(self, db_name):
        try:
            self.db_name = str(db_name)
            self.client = MongoClient(
                os.getenv('MONGO_URI', ul.MONGO_URI),
                # ssl_cert_reqs=ssl.CERT_NONE
            )
            self.db = self.client[str(db_name)]
        except Exception as e:
            logger.error('Error while connecting to db')
            logger.error(e)


    def get_collection(self, name):
        try:
            return self.db[name]
        except Exception as e:
            logger.error('Error while connecting to db')
            logger.error(e)


    def create_view(self, view_name, view_on, pipeline):
        return self.db.command({
            "create": view_name,
            "viewOn": view_on,
            "pipeline": pipeline
        })


    def execute_pipeline(self, name: str, pipeline: list):
        try:
            return list(self.db.get_collection(name).aggregate(pipeline))
        except Exception as e:
            logger.error('Error while executing pipeline')
            logger.error(e)


    def close_connection(self):
        try:
            if self.client:
                self.client.close()
        except Exception as e:
            logger.error('Error while closing connection to db')
            logger.error(e)


if __name__ == '__main__':
    mongo = Mongodb('11249')
    val = mongo.get_collection(mongo.GUEST_COLLECTION)
    print(val)
    mongo.close_connection()



from typing import Optional
from pydantic import BaseModel


class Country(BaseModel):
    orgId: str
    clientId: Optional[list] = []
    name: str
    statisticCode: str
    guestAddressFormat: str
    regionCode: str
    region: str
    isoCode: str
    displayFlag: bool
    countryCode: str



from typing import Optional
from pydantic import BaseModel


class Token(BaseModel):
    key: str
    refreshKey: Optional[str] = ''
    validity: str
    hostName: str
    appKey: Optional[str] = ''



from .property import PropertyModel, PropertyAddressModel
from .guest import ProfileIDObject, GuestsModel, GuestsModelV2
from .reservation import ReservationModel, Status, ReservationGuestInfo, ReservationGuestShort, ReservationModelV2


# from datetime import datetime
# from upswingutil.resource import get_model_from_cloud_storage
# from upswingutil.pms.rms import NAME, get_key, validate_key
# from upswingutil.db import Mongodb, Firestore
# from upswingutil.schema import Token
# import upswingutil as ul
# import requests
# import logging
# import asyncio
# from concurrent.futures import ThreadPoolExecutor
#
#
# def _get_geo_region(nationalID: int):
#     result = 'unknown'
#     try:
#         mongo = Mongodb('upswing')
#         region = mongo.get_collection('countries').find_one({"_id": nationalID}, {"region": 1})
#         mongo.close_connection()
#         result = region if region else 'unknown'
#     except Exception as e:
#         logging.error(f'Error getting geo region of nation {nationalID}')
#         logging.error(e)
#     finally:
#         return result
#
#
# @DeprecationWarning
# class ReservationSync:
#
#     def __init__(self, orgId: str, g_cloud_token=None):
#         self.orgId = orgId
#         self.mongo = Mongodb(orgId)
#         self.token: Token = get_key(self.orgId)
#         self.g_cloud_token = g_cloud_token
#         self.__urgent_booking_criteria__ = 5 * 86400
#         self.__stay_duration_criteria__ = 10 * 86400
#         self.__booking_category_levels__ = ['bronze', 'silver', 'gold', 'platinum']
#
#     def _get_booking_type(self, arrival_date: str, created_date: str):
#         result = 'unknown'
#         try:
#             delta = datetime.fromisoformat(arrival_date) - datetime.fromisoformat(created_date)
#             delta = delta.days * 86400 + delta.seconds
#             result = 'Urgent Booking' if delta < self.__urgent_booking_criteria__ else 'Pre Planned'
#         except Exception as e:
#             logging.error('Error while calculating booking type')
#             logging.error(e)
#         finally:
#             return result
#
#     def _get_duration_type(self, departure_date: str, arrival_date: str):
#         result = 'unknown'
#         try:
#             stay_d = datetime.fromisoformat(departure_date) - datetime.fromisoformat(arrival_date)
#             delta = stay_d.days * 86400 + stay_d.seconds
#             result = 'Short Stay' if delta < self.__stay_duration_criteria__ else 'Long Stay'
#         except Exception as e:
#             logging.error('Error calculating stay duration type')
#             logging.error(e)
#         finally:
#             return result
#
#     def _get_booking_level(self, orgId, clientId, departure_date: str, arrival_date: str, spend, booking_type):
#         result = 'unknown'
#         try:
#             stay_d = datetime.fromisoformat(departure_date) - datetime.fromisoformat(arrival_date)
#             booking_type_id = 1 if booking_type == 'Urgent Booking' else 0
#             model = get_model_from_cloud_storage(
#                 str(orgId),
#                 f'booking_categorization_{clientId}.pkl',
#                 token=self.g_cloud_token
#             )
#             prediction = model.predict([[stay_d.days, spend, booking_type_id]])
#             level = prediction[0] if len(prediction) > 0 else None
#             result = self.__booking_category_levels__[level] if level else 'unknown'
#         except Exception as e:
#             logging.error('Error calculating booking_level of reservation')
#             logging.error(e)
#         finally:
#             return result
#
#     def _retrieve_data(self, id, name, url):
#         result = None
#         try:
#             logging.debug(f'Extracting {name} for reservation: {id}')
#             header = {
#                 'Content-Type': 'application/json',
#                 'authtoken': self.token.key
#             }
#             response = requests.request("GET", url, headers=header)
#             if response.status_code == 200:
#                 result = response.json()
#             else:
#                 logging.error(f'Error getting {name} for reservation {id} due to status code {response.status_code}')
#                 logging.error(url.format(id))
#         except Exception as e:
#             logging.error(f' {name} failed due to {e}')
#         finally:
#             return result
#
#     def extract_reservation_details(self, record):
#         feature_list = [
#             {
#                 'name': 'holds',
#                 'url': 'holds'
#             },
#             {
#                 'name': 'guests',
#                 'url': 'guests'
#             },
#             {
#                 'name': 'billTo',
#                 'url': 'billTo'
#             },
#             {
#                 'name': 'transfers',
#                 'url': 'transfers'
#             },
#             {
#                 'name': 'auditTrail',
#                 'url': 'auditTrail'
#             },
#             {
#                 'name': 'daily_rates',
#                 'url': 'dailyRates'
#             },
#             {
#                 'name': 'rego_access',
#                 'url': 'regoAccess'
#             },
#             {
#                 'name': 'requirement',
#                 'url': 'requirements'
#             },
#             {
#                 'name': 'housekeeping',
#                 'url': 'housekeeping'
#             },
#             {
#                 'name': 'daily_revenue',
#                 'url': 'dailyRevenue'
#             },
#             {
#                 'name': 'add_ons',
#                 'url': 'reservationAddOn'
#             },
#             {
#                 'name': 'correspondence',
#                 'url': 'correspondence'
#             },
#             {
#                 'name': 'financial_info_actual',
#                 'url': 'actualAccount'
#             },
#             {
#                 'name': 'bedConfiguration',
#                 'url': 'bedConfiguration'
#             },
#         ]
#         try:
#             logging.info(f"Extracting reservation {record.get('id')} additional details.")
#
#             for item in feature_list:
#                 _url = f"{self.token.hostName}/reservations/{record.get('id')}/{item.get('url')}"
#                 record[item.get('name')] = self._retrieve_data(record.get("id"), item.get("name"), _url)
#             return record
#         except Exception as e:
#             logging.error(f'Error retrieving addition reservation details for {record.get("id")}')
#             logging.error(e)
#
#     def transform_reservation(self, record):
#         try:
#             logging.info(f'Transforming reservation {record.get("id")}')
#             record['booking_type'] = self._get_booking_type(record['arrivalDate'], record['createdDate'])
#             record['duration_type'] = self._get_duration_type(record['departureDate'], record['arrivalDate'])
#             record['globalId'] = f"{record.get('agent')}-{record.get('orgId')}-{record.get('clientId')}-{record.get('id')}"
#             record['booking_level'] = self._get_booking_level(
#                 record.get('orgId'),
#                 record['clientId'], record['departureDate'],
#                 record['arrivalDate'], record['financial_info_actual']['totalRate'],
#                 record['booking_type']
#             )
#             return record
#         except Exception as e:
#             logging.error(f"Exception while transforming reservation {record.get('id')}")
#             logging.error(e)
#
#     def transform_guest(self, record):
#         try:
#             logging.info(f'Transforming guests of reservation {record.get("id")}')
#             for guest in record.get('guests'):
#                 guest['_id'] = guest.get('id')
#                 del guest['id']
#                 guest['geoRegion'] = _get_geo_region(guest.get('nationalityId'))
#             return record
#         except Exception as e:
#             logging.error(f'Error while transforming guest for reservation {record.get("id")}')
#             logging.error(e)
#
#     def _add_reservation_to_alvie(self, record):
#         firestore_db = Firestore(app='alvie')
#         user_email_list = [item["email"] for item in record.get('guests')]
#         resv_info = {
#             "_id": record.get("id"),
#             "agent": record.get("agent"),
#             "orgId": record.get("orgId"),
#             "areaId": record.get("areaId"),
#             "arrivalDate": record.get("arrivalDate"),
#             "departureDate": record.get("departureDate"),
#             "propertyId": record.get("propertyId"),
#             "propertyName": record.get("propertyName"),
#             "status": record.get("status"),
#             "travelAgentId": record.get("travelAgentId"),
#             "travelAgentName": record.get("travelAgentName")
#         }
#         logging.debug(f"Final reservation id : {record.get('id')}")
#         for email in user_email_list:
#             docs = firestore_db.get_collection('users').where("email", "==", email).stream()
#             for doc in docs:
#                 logging.info("loading reservation to alvie")
#                 firestore_db.get_collection(f'users/{doc.id}/reservations') \
#                     .document(str(record.get("id"))) \
#                     .set(resv_info, merge=True)
#                 logging.debug(f"added to {doc.id}")
#
#     def extract_reservation(self, record):
#         reservationId = record.get("reservation")
#         logging.info(f'Extracting reservation: {reservationId}')
#
#         header = {
#             'Content-Type': 'application/json',
#             'authtoken': self.token.key
#         }
#         url = f"{self.token.hostName}/reservations/{reservationId}?modelType=full"
#         response = requests.request("GET", url, headers=header)
#         if response.status_code == 200:
#             try:
#                 response_json = dict(response.json())
#                 response_json['orgId'] = self.orgId
#                 response_json['agent'] = record.get('agent')
#                 property_info = self.mongo.get_collection(self.mongo.PROPERTY_COLLECTION).find_one(
#                     {"areas.id": response_json.get('areaId')},
#                     {'id': 1, 'name': 1,
#                      'clientId': 1}
#                 )
#                 response_json['clientId'] = property_info.get('clientId') if property_info else 0
#                 response_json['propertyId'] = property_info.get('_id') if property_info else 0
#                 response_json['propertyName'] = property_info.get('name') if property_info else 'Unknown'
#                 return response_json
#             except Exception as err:
#                 logging.error(f'{record} load failed due to {err}')
#         else:
#             logging.error(f"RMS returned status code {response.status_code} "
#                           f"for reservation {reservationId} ")
#             raise Exception('Reservation not found / unable to retrieve')
#
#     def _load_to_mongodb(self, record):
#         _id = record.get("id")
#         orgId = record.get("orgId")
#         record.update({'lastUpdate': str(datetime.now())})
#         try:
#             logging.info(f'Adding reservation {_id} to db {orgId}')
#             self.mongo.get_collection(self.mongo.RESERVATION_COLLECTION).update_one({'_id': _id},
#                                                                           {'$set': record}, upsert=True)
#         except Exception as e:
#             logging.error(f'Error storing to respective {_id} to db {orgId}')
#             logging.error(e)
#
#     def process(self, reservationId):
#         record = {
#             'orgId': self.orgId,
#             'agent': NAME,
#             'reservation': reservationId
#         }
#
#         if not validate_key(self.token.validity):
#             logging.info('Refreshing RMS token as it is about to expire')
#             self.token = get_key(self.orgId)
#
#         record = self.extract_reservation(record)
#         record = self.extract_reservation_details(record)
#         record = self.transform_reservation(record)
#         record = self.transform_guest(record)
#         self._load_to_mongodb(record)
#         self._add_reservation_to_alvie(record)
#
#     def __del__(self):
#         if self.mongo:
#             self.mongo.close_connection()
#
#
# if __name__ == '__main__':
#     ul.G_CLOUD_PROJECT = 'aura-staging-31cae'
#     ul.ENCRYPTION_SECRET = "S1335HwpKYqEk9CM0I2hFX3oXa5T2oU86OXgMSW4s6U="
#     ul.FIREBASE = '/Users/harsh/upswing/github/agent-oracle/SECRET/aura-staging-31cae-firebase-adminsdk-dyolr-' \
#                   '7c135838e9.json'
#     ul.MONGO_URI = "mongodb://AdminUpSwingGlobal:Upswing098812Admin0165r@dev.db.upswing.global:27017/?authSo" \
#                    "urce=admin&readPreference=primary&appname=Agent%20RMS%20Dev&ssl=false"
#     import firebase_admin
#     cred = firebase_admin.credentials.Certificate(ul.FIREBASE)
#     firebase = firebase_admin.initialize_app(cred)
#     resv = ReservationSync('11249')
#     resv.process(23560)
#     resv.process(23590)



from upswingutil.resource import decrypt, http_retry
from loguru import logger
from upswingutil.schema import Token
from datetime import datetime, timedelta
from upswingutil.db import Firestore


def validate_key(expiryDate):
    try:
        if expiryDate is None:
            return False
        return datetime.fromisoformat(expiryDate) > (datetime.now() + timedelta(minutes=30))
    except Exception as e:
        logger.error('Error while validating key')
        logger.error(e)
        return False


def get_key(org_id):
    db = Firestore()
    client = db.get_collection(db.org_collection).document(org_id).get().to_dict()
    url = client["pms"]["url"]
    body = {
        "agentId": client["pms"]["agentId"],
        "agentPassword": decrypt(client["pms"]["agentPassword"]),
        "clientId": client["pms"]["clientId"],
        "clientPassword": decrypt(client["pms"]["clientPassword"]),
        "useTrainingDatabase": client["pms"]["useTrainingDatabase"],
        "moduleType": client["pms"]["moduleType"]
    }
    try:
        http = http_retry()
        resp = http.post(url + '/authToken', json=body)
        if resp.status_code != 200:
            logger.error('POST /authToken {}'.format(resp.status_code))
            logger.error(resp.__str__())
            logger.error(resp.reason)
        else:
            return Token(
                key=resp.json()['token'],
                validity=resp.json()['expiryDate'],
                hostName=url
            )
    except Exception as e:
        logger.error('Exception generating auth token for RMS')
        logger.error(e)



import upswingutil as ul
from upswingutil.pms.oracle import ReservationSync
import firebase_admin
from loguru import logger
import os

ul.ENCRYPTION_SECRET = "S1335HwpKYqEk9CM0I2hFX3oXa5T2oU86OXgMSW4s6U="
ul.MONGO_URI = "mongodb://AdminUpSwingGlobal:Upswing098812Admin0165r@dev.db.upswing.global:27017/?authSource=admin&readPreference=primary&appname=Agent%20Oracle%20Dev&ssl=false"
ul.G_CLOUD_PROJECT = "aura-staging-31cae"
ul.FIREBASE = "SECRET/aura-staging-31cae-firebase-adminsdk-dyolr-7c135838e9.json"
ul.LOG_LEVEL_VALUE = 'DEBUG'


try:
    firebase_admin.get_app()
except Exception as e:
    logger.info(e)
    logger.info('Initializing default firebase app')
    cred = firebase_admin.credentials.Certificate(
        os.getenv("FIREBASE", "/Users/harsh/upswing/github/api-oracle/SECRET/aura-staging-31cae-firebase-adminsdk-dyolr-7c135838e9.json"))
    firebase = firebase_admin.initialize_app(cred)


if __name__ == '__main__':
    print('testing resv sync')
    orgId = 'OHIPSB'
    resvId = '20908'
    hotelId = 'SAND01'
    resv = ReservationSync(orgId)
    resv.process(resvId, hotelId)
    print('completed')



import logging
from upswingutil.db import Firestore
from upswingutil.db.model import ReservationModelV2


def store_reservation_to_alvie(record: ReservationModelV2):
    try:
        alvieDB = Firestore(app='alvie')
        record = record.to_mongo()
        for guest in record.get('guestInfo').get('guest_list'):
            guest = dict(guest)
            # to store reservation only in the profile of the primary guest
            # incase if we have to make it store for all the guests, then remove this condition
            if guest.get('primary'):
                docs = alvieDB.get_collection(f'Organizations/{record.get("orgId")}/users') \
                    .where("pmsGuestProfile", "==", guest.get('guest')).stream()
                for doc in docs:
                    alvieDB.get_collection(f'Organizations/{record.get("orgId")}/users/{doc.id}/reservations') \
                        .document(str(record.get('_id'))).set(record, merge=True)
                    logging.info(f"reservation {record.get('orgId')} : {record.get('_id')} added to alvie profile {doc.id}")
    except Exception as e:
        logging.error(f'Error while storing reservation {record.get("orgId")} : {record.get("_id")} to firestore')
        logging.error(e)



from upswingutil.pms import store_reservation_to_alvie
from upswingutil.pms.oracle import NAME, get_key, validate_key
from upswingutil.db import Mongodb, MongodbV2
from upswingutil.schema import Token
from datetime import datetime, timedelta
import requests
import logging
from upswingutil.db.model import ReservationModelV2, Status, ReservationGuestInfo, ReservationGuestShort, \
    GuestsModelV2


def _get_status(status):
    if status == 'NoShow':
        return Status.NO_SHOW
    elif status == 'CheckedOut' or status == 'Departed':
        return Status.DEPARTED
    elif status == 'Reserved':
        return Status.RESERVED
    elif status == 'Cancelled':
        return Status.CANCELLED
    elif status == 'InHouse':
        return Status.IN_HOUSE
    else:
        logging.error(f'Missing Status : {status}')
        return Status.UNCONFIRMED


def _get_id(record):
    _id = next((x['id'] for x in record.get('reservationIdList') if x['type'] == 'Reservation'), None)
    return _id


def _get_transaction_info(transactions):
    return transactions


def _get_package_info(package):
    return [{
        'packageCode': item.get('packageCode'),
        'ratePlanCode': item.get('ratePlanCode'),
        'source': item.get('source'),
        'description': item.get('packageHeaderType').get('primaryDetails').get('description'),
        'shortDescription': item.get('packageHeaderType').get('primaryDetails').get('shortDescription'),
        'amount': sum([x['unitPrice'] for x in item.get('scheduleList')]) * item.get('consumptionDetails').get(
            'totalQuantity')
    } for item in package] if package else []


def _generate_day_to_day_entry(record):
    _daily_activity = list()
    start_date = datetime.fromisoformat(record.get('roomStay').get('arrivalDate'))
    end_date = datetime.fromisoformat(record.get('roomStay').get('departureDate'))
    day_delta = timedelta(days=1)
    for i in range((end_date - start_date).days):
        _rates = next((x for x in record.get('roomStay').get('roomRates') if
                       x['start'] == str((start_date + i * day_delta).date())), None)
        _daily_activity.append({
            'date': str((start_date + i * day_delta).date()),
            'day_of_stay': i + 1,
            'roomType': _rates.get('roomType'),
            'ratePlanCode': _rates.get('ratePlanCode'),
            'rateTypeId': _rates.get('ratePlanCode'),
            'roomId': _rates.get('roomId'),
            'roomName': _rates.get('roomId'),
            'rates': {
                'currency': _rates.get('rates').get('rate')[0].get('base').get('currencyCode'),
                'amountBeforeTax': _rates.get('rates').get('rate')[0].get('base').get('amountBeforeTax'),
                'discountAmount': None,
                'packageAmount': None,
                'totalRateAmount': _rates.get('total').get('amountBeforeTax')
            },
            'revenue': {
                'accommodation': (
                            record.get('cashiering').get('revenuesAndBalances').get('roomRevenue').get('amount') / (
                                end_date - start_date).days),
                'accommodationTax': 0,
                'accommodationGST': 0,
                'foodAndBeverage': (record.get('cashiering').get('revenuesAndBalances').get('foodAndBevRevenue').get(
                    'amount') / (end_date - start_date).days),
                'foodAndBeverageTax': 0,
                'foodAndBeverageGST': 0,
                'other': 0,
                'otherTax': 0,
                'otherGST': 0
            },
            'guestCounts': {
                'adults': _rates.get('guestCounts').get('adults'),
                'children': _rates.get('guestCounts').get('children'),
                'infants': 0
            }
        })
    return _daily_activity


def _create_reservation_id_obj(record) -> dict:
    return {
        'reservation': str(record.get('id')),
        'globalId': f"{record.get('agent')}-{record.get('orgId')}-{record.get('clientId')}-{record.get('id')}",
        'onlineConfirmationId': str(record.get('idObj').get('confirmation')),
        'hotelId': str(record.get('hotelId')),
        'createdById': str(record.get('creatorId')),
        'modifiedById': str(record.get('lastModifierId')),
        'roomId': str(record.get('roomStay').get('roomRates')[0].get('roomId'))
    }


def _transform_cancellation_record(record, policy) -> list:
    _cancellation_list = list()
    for item in record:
        _cancellation_list.append({
            'reason': item.get('reason'),
            'cxlDate': item.get('cxlDate'),
            'userId': item.get('userId'),
            'policy': policy,
            'externalCancelId': None,
            'businessLostId': None
        })
    return _cancellation_list


def _create_guest_id_obj(record):
    return {
        'profile': str((next(x for x in record.get('profileIdList') if x['type'] == 'Profile')).get('id')),
        'corporateId': str((next(x for x in record.get('profileIdList') if x['type'] == 'CorporateId')).get('id'))
    }


def _get_contact_details(record):
    _data = dict()

    if record.get('emails'):
        _data['email'] = next((x['email']['emailAddress'] for x in record.get('emails').get('emailInfo') if
                               x['email']['primaryInd'] is True), '')
        _data['email2'] = next((x['email']['emailAddress'] for x in record.get('emails').get('emailInfo') if
                                x['email']['primaryInd'] is False), '')
    else:
        _data['email'] = ''
        _data['email2'] = ''

    if record.get('telephones'):
        _data['mobile'] = next((x['telephone']['phoneNumber'] for x in record.get('telephones').get('telephoneInfo') if
                                x['telephone']['phoneUseType'] == 'MOBILE'), '')
        _data['telephone'] = next(
            (x['telephone']['phoneNumber'] for x in record.get('telephones').get('telephoneInfo') if
             x['telephone']['phoneUseType'] == 'BUSINESS'), '')
    else:
        _data['mobile'] = ''
        _data['telephone'] = ''

    return _data


def _get_address_details(record):
    _address1 = ''
    _address2 = ''
    _address3 = ''
    _state = ''
    _town = ''
    _postCode = ''
    if record.get('addresses').get('addressInfo') and len(record.get('addresses').get('addressInfo')) > 0:
        _address1 = record.get('addresses').get('addressInfo')[0].get('address').get('addressLine')[0] if len(
            record.get('addresses').get('addressInfo')[0].get('address').get('addressLine')) >= 1 else ''
        _address2 = record.get('addresses').get('addressInfo')[0].get('address').get('addressLine')[1] if len(
            record.get('addresses').get('addressInfo')[0].get('address').get('addressLine')) >= 2 else ''
        _address3 = record.get('addresses').get('addressInfo')[0].get('address').get('addressLine')[2] if len(
            record.get('addresses').get('addressInfo')[0].get('address').get('addressLine')) >= 3 else ''
        _state = record.get('addresses').get('addressInfo')[0].get('state')
        _town = record.get('addresses').get('addressInfo')[0].get('cityName')
        _postCode = record.get('addresses').get('addressInfo')[0].get('postalCode')

    return {
        'addressLine1': _address1,
        'addressLine2': _address2,
        'addressLine3': _address3,
        'countryId': record.get('customer').get('citizenCountry').get('code') if record.get('customer') else '',
        'nationalityId': record.get('customer').get('citizenCountry').get('code') if record.get('customer') else '',
        'state': _state,
        'town': _town,
        'postCode': _postCode
    }


def _transform_guest_profile(record):
    idObj = _create_guest_id_obj(record)
    bday = record.get('birthDate')
    record = record.get('profileDetails')
    guest = GuestsModelV2(
        _id=idObj.get('profile'),
        idObj=idObj,
        profileType=record.get('profileType'),
        title='',
        firstName=record.get('customer').get('personName')[0].get('givenName') if record.get(
            'profileType') == 'Guest' else record.get('company').get('companyName'),
        lastName=record.get('customer').get('personName')[0].get('surname') if record.get(
            'profileType') == 'Guest' else '',
        datesAndDurations={
            'birthDate': bday,
            'anniversary': None
        },
        address=_get_address_details(record),
        contactInfo=_get_contact_details(record),
        privacyInfo={
            'emailOptOut': not record.get('privacyInfo').get('optInEmail'),
            'marketingOptOut': record.get('privacyInfo').get('optInMarketResearch'),
            'smsOptOut': not record.get('privacyInfo').get('optInSms'),
            'privacyOptIn': record.get('privacyInfo').get('infoFromThirdParty'),
            'phoneOptOut': not record.get('privacyInfo').get('optInPhone')
        },
        documents={
        },
        metaInfo={
            'uRLs': record.get('uRLs'),
            'notes': record.get('comments'),
            'languageSpokenId': record.get('customer').get('language') if record.get('customer') else '',
            'profileDeliveryMethods': record.get('profileDeliveryMethods'),
            'profileMemberships': record.get('profileMemberships'),
            'preferenceCollection': record.get('preferenceCollection'),
            'keywords': record.get('keywords'),
            'profileIndicators': record.get('profileIndicators'),
            'relationships': record.get('relationships'),
            'relationshipsSummary': record.get('relationshipsSummary'),
            'profileAccessType': record.get('profileAccessType'),
            'profileRestrictions': record.get('profileRestrictions'),
            'taxInfo': record.get('taxInfo'),
            'salesInfo': record.get('salesInfo'),
            'subscriptions': record.get('subscriptions'),
        },
        company=record.get('company'),
        registeredProperty=str(record.get('registeredProperty')),
        auraRecordUpdatedOn=datetime.now(),
    )
    return guest


class ReservationSync:

    def __init__(self, orgId: str, g_cloud_token=None):
        self.orgId = orgId
        self.mongo = Mongodb(orgId)
        self.mongoV2 = MongodbV2(orgId)
        self.token: Token = get_key(self.orgId)
        self.g_cloud_token = g_cloud_token
        self._api_call_counter = 0

    def _extract_guest_profile(self, hotelId, guest):
        try:
            guestId = 0
            for item in guest.get('profileInfo').get('profileIdList'):
                if item.get('type') == 'Profile':
                    guestId = item.get('id')

            if not validate_key(self.token.validity):
                logging.info(f'Refreshing {NAME} token as it is about to expire')
                self.token = get_key(self.orgId, self.token.refreshKey)

            _url = f'{self.token.hostName}/crm/v1/profiles/{guestId}?fetchInstructions=Address' \
                   f'&fetchInstructions=Comment&fetchInstructions=Communication&fetchInstructions=Correspondence' \
                   f'&fetchInstructions=DeliveryMethods&fetchInstructions=FutureReservation' \
                   f'&fetchInstructions=GdsNegotiatedRate&fetchInstructions=HistoryReservation' \
                   f'&fetchInstructions=Indicators&fetchInstructions=Keyword&fetchInstructions=Membership' \
                   f'&fetchInstructions=NegotiatedRate&fetchInstructions=Preference&fetchInstructions=Profile' \
                   f'&fetchInstructions=Relationship&fetchInstructions=SalesInfo&fetchInstructions=Subscriptions' \
                   f'&fetchInstructions=WebUserAccount'
            _profile_data = self.retrieve_data(hotelId, _url)
            if _profile_data:
                idObj = dict()
                for item in _profile_data['profileIdList']:
                    idObj[item.get('type').lower()] = item['id']
                _profile_data['idObj'] = idObj
                _profile_data['birthDate'] = guest.get('birthDate')
                return _profile_data
            else:
                logging.error(f'Guest profile not found : {guestId}')

        except Exception as e:
            logging.error(f'Error while transforming guest {guest}')
            logging.error(e)

    def _get_guest_list(self, hotelId, record):
        logging.info(f'Processing guest list of size: {len(record.get("reservationGuests"))}')
        _guest_list = list()
        for item in record.get('reservationGuests'):
            _guest_profile = _transform_guest_profile(self._extract_guest_profile(hotelId, item))
            self.mongoV2.save(_guest_profile)
            _guest_list.append(ReservationGuestShort(
                guest=_guest_profile.id,
                primary=item.get('primary'),
                birthDate=item.get('birthDate'),
                arrivalTransport=item.get('arrivalTransport'),
                departureTransport=item.get('departureTransport')
            ))
        return _guest_list

    def _get_guest_info(self, record):
        guest_info = ReservationGuestInfo(
            adults=record.get('roomStay').get('guestCounts').get('adults'),
            children=record.get('roomStay').get('guestCounts').get('children'),
            infants=0,
            childBuckets=record.get('roomStay').get('guestCounts').get('childBuckets'),
            preRegistered=record.get('preRegistered'),
            guest_list=self._get_guest_list(record.get('hotelId'), record)
        )
        return guest_info

    def _get_reservation_id(self, idList, agent, hotelId, resv):
        resv_dict = dict()
        resv_dict['globalId'] = f'{agent}-{self.orgId}-{hotelId}-{resv}'
        for item in idList:
            value = item['id']
            _type = item['type']
            resv_dict[_type.lower()] = value
        return resv_dict

    def retrieve_data(self, clientId, url, payload="") -> dict:
        headers = {
            'Content-Type': 'application/json',
            'x-hotelid': clientId,
            'x-app-key': self.token.appKey,
            'Authorization': f'Bearer {self.token.key}'
        }
        response = requests.request("GET", url, headers=headers, data=payload)
        self._api_call_counter += 1
        if response.status_code == 200:
            return dict(response.json())
        else:
            logging.error(f'Failed to retrieve data, status: {response.status_code}')
            logging.error(response.reason)
            return None

    def _extract_reservation_details(self, record):
        resv_id = record.get('id')
        try:
            hotel_id = record.get('hotelId')
            record['idObj'] = self._get_reservation_id(record.get('reservationIdList'), record.get('agent'),
                                                       record.get('hotelId'), record.get('id'))
            logging.info(f"Extracting reservation {resv_id} additional details.")
            if self.token is None:
                self.token: Token = record.get('token')

            if not validate_key(self.token.validity):
                logging.info(f'Refreshing {NAME} token as it is about to expire')
                self.token = get_key(self.orgId, self.token.refreshKey)

            _url = f'{self.token.hostName}/rsv/v0/hotels/{hotel_id}/reservations/{resv_id}/cancellationHistory'
            data = self.retrieve_data(record.get('hotelId'), _url)
            record['cancellation'] = data.get('cxlActivityLog') if data else None

            _url = f'{self.token.hostName}/rsv/v0/hotels/{hotel_id}/reservations/{resv_id}/linkedSummary'
            data = self.retrieve_data(record.get('hotelId'), _url)
            record['linkedReservation'] = data.get('linkedReservationList') if data else None

            # _url = f'{self.token.hostName}/rsv/v0/hotels/{hotel_id}/reservations/{resv_id}/calls'
            # data = self.retrieve_data(record.get('hotelId'), _url)
            # record['callHistory'] = data.get('callHistory') if data else None

            # _url = f'{self.token.hostName}/rsv/v0/hotels/{hotel_id}/reservations/{resv_id}/inventoryItems'
            # data = self.retrieve_data(record.get('hotelId'), _url)
            # record['inventoryItems'] = data.get('inventoryItems') if data else None

            _url = f'{self.token.hostName}/csh/v0/hotels/{hotel_id}/reservations/{resv_id}/folios?includeFolioHistory=true&fetchInstructions=Account&fetchInstructions=Payee&fetchInstructions=Totalbalance&fetchInstructions=Windowbalances&fetchInstructions=Payment&fetchInstructions=Postings&fetchInstructions=Transactioncodes&fetchInstructions=Reservation'
            data = self.retrieve_data(record.get('hotelId'), _url)
            record['folioInformation'] = data.get('reservationFolioInformation') if data else None

            # _url = f'{self.token.hostName}/csh/v0/hotels/{hotel_id}/transactions?reservationList={resv_id}&idContext=OPERA&type=Reservation&includeGenerates=true&includeTransactionsWithFolioNo=true&includeTransactionsWithManualPostingOnly=true'
            # data = self.retrieve_data(record.get('hotelId'), _url)
            # if data:
            #     del data['links']
            # record['transactions'] = data

            return record
        except Exception as e:
            logging.error(f'Error retrieving addition reservation details for {resv_id}')
            logging.error(e)

    def _extract_reservation(self, record):
        try:
            reservationId = record.get("reservation")["reservation"]
            hotelId = record.get('reservation')['hotelId']
            logging.info(f'Extracting reservation: {reservationId}')

            header = {
                'Content-Type': 'application/json',
                'x-hotelid': hotelId,
                'x-app-key': self.token.appKey,
                'Authorization': f'Bearer {self.token.key}'
            }

            url = f"{self.token.hostName}/rsv/v1/hotels/{hotelId}/reservations/{reservationId}?fetchInstructions" \
                  f"=Comments&fetchInstructions=GuestMemberships&fetchInstructions=GuestLastStay&fetchInstructions" \
                  f"=ProfileAwards&fetchInstructions=ScheduledActivities&fetchInstructions=ServiceRequests" \
                  f"&fetchInstructions=ReservationAwards&fetchInstructions=RevenuesAndBalances&fetchInstructions" \
                  f"=Tickets&fetchInstructions=GuestComments&fetchInstructions=Packages&fetchInstructions" \
                  f"=InventoryItems&fetchInstructions=ReservationPaymentMethods&fetchInstructions=RoutingInstructions" \
                  f"&fetchInstructions=Preferences&fetchInstructions=Memberships&fetchInstructions=Alerts" \
                  f"&fetchInstructions=Traces&fetchInstructions=ConfirmationLetters&fetchInstructions=CallHistory" \
                  f"&fetchInstructions=FixedCharges&fetchInstructions=GuestMessages&fetchInstructions" \
                  f"=ReservationPolicies&fetchInstructions=Indicators&fetchInstructions=LinkedReservations" \
                  f"&fetchInstructions=ECoupons&fetchInstructions=TrackItItems&fetchInstructions=WebRegistrationCards" \
                  f"&fetchInstructions=ServiceRequests&fetchInstructions=ReservationActivities&fetchInstructions" \
                  f"=PrepaidCards&fetchInstructions=Shares&fetchInstructions=Attachments&fetchInstructions=Locators" \
                  f"&fetchInstructions=TransactionDiversions&fetchInstructions=ECertificates&fetchInstructions" \
                  f"=UpsellInfo&fetchInstructions=RoomAssignedByAI&fetchInstructions=Reservation"
            response = requests.request("GET", url, headers=header)
            self._api_call_counter += 1
            if response.status_code == 200:
                response_json = dict(response.json()).get('reservations').get('reservation')
                if len(response_json) > 0:
                    result = response_json[0]
                    result['id'] = reservationId
                    result['token'] = self.token
                    result['orgId'] = record.get('orgId')
                    result['agent'] = record.get('agent')
                    return result
                else:
                    logging.error(f"{NAME} returned reservation {reservationId} of length zero")
            else:
                logging.error(f"{NAME} returned status code {response.status_code} for reservation {reservationId} ")
        except Exception as err:
            logging.error(f'{record} load failed due to {err}')
        return None

    def _transform_reservation(self, record):
        reservationV2 = ReservationModelV2(
            _id=str(record.get('id')),
            orgId=self.orgId,
            agent=record.get('agent'),
            idObj=_create_reservation_id_obj(record),
            datesAndDuration={
                'createdDate': record.get('createDateTime'),
                'confirmedDate': record.get('createDateTime'),
                'arrivalDate': record.get('roomStay').get('arrivalDate'),
                'departureDate': record.get('roomStay').get('departureDate'),
                'cancelledDate': record.get('cancellation')[0].get('cxlDate'),
                'expectedTimes': {
                    'arrival': record.get('roomStay').get('expectedTimes').get('reservationExpectedArrivalTime'),
                    'departure': record.get('roomStay').get('expectedTimes').get('reservationExpectedDepartureTime')
                },
                'originalTimeSpan': None,
                'lastModifyDateTime': record.get('lastModifyDateTime'),
                'createBusinessDate': record.get('createBusinessDate'),
                'createDateTime': record.get('createDateTime'),
                'eta': None,
                'event': None
            },
            status=_get_status(record.get('reservationStatus')),
            correspondence=None,
            metaInfo={
                'allowMobileCheckout': record.get('allowMobileCheckout'),
                'allowMobileViewFolio': record.get('allowMobileViewFolio'),
                'allowPreRegistration': record.get('allowPreRegistration'),
            },
            property={
                'hotelId': str(record.get('hotelId')),
                'hotelName': str(record.get('hotelId')),
                'roomId': str(record.get('folioInformation').get('reservationInfo').get('roomStay').get('roomId')),
                'roomName': str(record.get('folioInformation').get('reservationInfo').get('roomStay').get('roomId')),
                'categoryName': record.get('folioInformation').get('reservationInfo').get('roomStay').get('roomType'),
                'roomClass': record.get('folioInformation').get('reservationInfo').get('roomStay').get('roomClass'),
                'numberOfRooms': record.get('folioInformation').get('reservationInfo').get('roomStay').get(
                    'numberOfRooms'),
                'bedConfiguration': None
            },
            stayInfo={
                'longTerm': None,
                'businessSegmentId': None
            },
            reservationTypes={
                'type': 'Normal'
            },
            bookingInfo={
                'upgradeEligible': record.get('upgradeEligible'),
                'upgradeReason': None,
                'bookingMedium': record.get('roomStay').get('bookingMedium'),
                'bookingMediumDescription': record.get('roomStay').get('bookingMediumDescription'),
                'sourceOfSaleType': record.get('sourceOfSale').get('sourceType'),
                'sourceOfSaleCode': record.get('sourceOfSale').get('sourceCode'),
                'travelAgentId': record.get('folioInformation').get('reservationInfo').get('roomStay').get(
                    'bookingChannelCode'),
                'travelAgentName': record.get('folioInformation').get('reservationInfo').get('roomStay').get(
                    'bookingChannelCode'),
                'marketSegment': record.get('roomStay').get('roomRates')[0].get('marketCodeDescription'),
                'subMarketSegment': None,
                'notes': None,
            },
            financeInfo={
                'accountId': None,
                'rateTypeId': record.get('folioInformation').get('reservationInfo').get('roomStay').get('ratePlanCode'),
                'rateTypeName': record.get('folioInformation').get('reservationInfo').get('roomStay').get(
                    'ratePlanCode'),
                'totalPoints': record.get('roomStay').get('totalPoints').get('points'),
                'totalSpending': {
                    'amountBeforeTax': record.get('cashiering').get('revenuesAndBalances').get('totalRevenue').get(
                        'amount'),
                    'taxAmount': record.get('cashiering').get('revenuesAndBalances').get('totalPayment').get(
                        'amount') - record.get('cashiering').get('revenuesAndBalances').get('totalRevenue').get(
                        'amount') if (record.get('cashiering').get('revenuesAndBalances').get('totalPayment').get(
                        'amount') and record.get('cashiering').get('revenuesAndBalances').get('totalRevenue').get(
                        'amount')) else 0,
                    'totalRate': record.get('cashiering').get('revenuesAndBalances').get('totalPayment').get('amount')
                },
                'projectedTotalSpending': None,
                'paymentMethod': record.get('reservationPaymentMethods'),
                'other': None,
                'revenueBucketsInfo': None,
                'transactions': [],
                'revenue': {
                    'totalFixedCharge': {
                        'amount': record.get('cashiering').get('revenuesAndBalances').get('totalRevenue').get('amount')
                    },
                    'totalPayment': {
                        'amount': record.get('cashiering').get('revenuesAndBalances').get('totalPayment').get('amount')
                    },
                    'roomRevenue': {
                        'amount': record.get('cashiering').get('revenuesAndBalances').get('roomRevenue').get('amount')
                    },
                    'foodAndBevRevenue': {
                        'amount': record.get('cashiering').get('revenuesAndBalances').get('foodAndBevRevenue').get(
                            'amount')
                    },
                    'otherRevenue': {
                        'amount': 0
                    },
                    'totalRevenue': {
                        'amount': record.get('cashiering').get('revenuesAndBalances').get('totalRevenue').get('amount')
                    },
                    'balance': {
                        'amount': record.get('cashiering').get('revenuesAndBalances').get('balance').get('amount')
                    }
                },
                'deposit': {
                    'deposit': record.get('folioInformation').get('reservationInfo').get('roomStay').get(
                        'depositPayments').get('amount') if record.get('folioInformation').get('reservationInfo').get(
                        'roomStay').get('depositPayments') else 0,
                    'depositRequiredByDate': record.get('reservationPolicies').get('depositPolicies')[0].get(
                        'policy').get('deadline').get('absoluteDeadline') if len(
                        record.get('reservationPolicies').get('depositPolicies')) > 0 else None,
                    'secondDeposit': 0,
                    'secondDepositRequiredByDate': None,
                },
                'discount': [],
                'package': _get_package_info(record.get('reservationPackages')),
                'commission': {
                    'travelAgentCommissionPercentage': None
                }
            },
            daily_activity=_generate_day_to_day_entry(record),
            linkedReservation=[record.get('linkedReservation')],
            guestInfo=self._get_guest_info(record),
            eCertificates=record.get('eCertificates'),
            historyEvents=[],
            cancellation=_transform_cancellation_record(record.get('cancellation'),
                                                        record.get('reservationPolicies').get('cancellationPolicies')),
            housekeeping=[],
            comments=[],
            policies=record.get('reservationPolicies'),
            inventoryItems=record.get('inventoryItems'),
            preferences=[],
            requirement=[],
            rego_access=None,
            memberships=record.get('reservationMemberships'),
            packages=record.get('reservationPackages'),
            transfers=None,
            auraRecordUpdatedOn=datetime.now()
        )

        return reservationV2

    def _load_to_mongodb(self, record):

        try:
            logging.info(f'Adding reservation {record.id} to db {record.orgId}')
            self.mongo.get_collection(self.mongo.RESERVATION_COLLECTION).update_one({'_id': record.id},
                                                                                    {'$set': record.dict(
                                                                                        exclude={'token', 'id'})},
                                                                                    upsert=True)
        except Exception as e:
            logging.error(f'Error storing to respective {record.id} to db {record.orgId}')
            logging.error(e)

    def process(self, reservationId, hotelId):
        record = {
            'orgId': self.orgId,
            'agent': NAME,
            'reservation': {
                'reservation': reservationId,
                'hotelId': hotelId
            }
        }

        if not validate_key(self.token.validity):
            logging.info(f'Refreshing {NAME} token as it is about to expire')
            self.token = get_key(self.orgId, self.token.refreshKey)

        record = self._extract_reservation(record)
        if record:
            del record['token']
            record = self._extract_reservation_details(record)
            record = self._transform_reservation(record)
            self.mongoV2.save(record)
            store_reservation_to_alvie(record)

    def __del__(self):
        logging.info(f'API calls made: {self._api_call_counter}')
        if self.mongo:
            self.mongo.close_connection()
        if self.mongoV2:
            self.mongoV2.close_connection()



import urllib
import requests
from datetime import datetime, timedelta
from upswingutil.db import Firestore
from upswingutil.resource import decrypt
from upswingutil.schema import Token
from loguru import logger


def validate_key(expiryDate):
    """
    Validated whether the token is still valid or not
    :param expiryDate:
    :return:
    """
    try:
        if expiryDate is None:
            return False
        return datetime.fromisoformat(expiryDate) > (datetime.now() + timedelta(minutes=5))
    except Exception as e:
        logger.error('Error while validating key')
        logger.error(e)
        return False


def get_key(org_id: str, refresh_toke=None) -> Token:
    """
    This method generates auth token for provided org_id, in case of refreshing token,
    provide refresh_token, to generate new token using it.

    :param org_id:
    :param refresh_toke:
    :return: instance of Token
    """
    try:
        db = Firestore()
        pms = db.get_collection(db.org_collection).document(org_id).get().to_dict().get('pms')
        url = f"{pms.get('hostName')}/oauth/v1/tokens"
        auth_base64 = pms.get('client').decode("utf-8")
        app_key = decrypt(pms.get('appKey'))
        if refresh_toke:
            payload = f'refresh_token={refresh_toke}&grant_type=refresh_token'
            logger.debug('Refreshing token key')
        else:
            logger.debug('Generating NEW token key')
            payload = urllib.parse.urlencode({
                'username': decrypt(pms.get('username')),
                'password': decrypt(pms.get('password')),
                'grant_type': 'password'
            })
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'x-app-key': app_key,
            'Authorization': f'Basic {auth_base64}'
        }
        response = requests.request("POST", url, headers=headers, data=payload)
        if response.status_code != 200:
            logger.error('POST /authToken {}'.format(response.status_code))
            logger.error(response.__str__())
            logger.error(response.reason)
        else:
            data = dict(response.json())
            return Token(
                key=data.get('access_token'),
                refreshKey=data.get('refresh_token'),
                validity=str(datetime.now() + timedelta(seconds=int(data.get('expires_in')))),
                hostName=pms.get('hostName'),
                appKey=app_key
            )
    except Exception as e:
        logger.error('Error generating token key oracle')
        logger.error(e)


if __name__ == '__main__':
    orgId = "SAND01"
    token, valid, refresh = get_key(orgId)
    print('with username password')
    print('token', token)
    print('valid', valid)
    print('refresh', refresh)
    print('with refresh token')
    token, valid, refresh = get_key(orgId, refresh)
    print('token', token)
    print('valid', valid)
    print('refresh', refresh)



from cryptography.fernet import Fernet
import os
import upswingutil as ul


def generate_key():
    """
    Generates a key and save it into a file
    """
    key = Fernet.generate_key()
    with open("secret.key", "wb") as key_file:
        key_file.write(key)


def _load_key():
    """
    Load the previously generated key
    """
    # return open("secret.key", "rb").read()
    return os.getenv('ENCRYPTION_SECRET', ul.ENCRYPTION_SECRET)


def encrypt(message):
    """
    Encrypts a message
    """
    key = _load_key()
    encoded_message = message.encode()
    f = Fernet(key)
    encrypted_message = f.encrypt(encoded_message)
    return encrypted_message


def decrypt(message):
    key = _load_key()
    f = Fernet(key)
    return f.decrypt(message).decode()


if __name__ == "__main__":
    val = encrypt("encrypt this message")
    print(decrypt(val))



from typing import Optional
from pydantic import BaseModel


class AlvieRating(BaseModel):
    overall: Optional[int] = 0
    foodAndBev: Optional[int] = 0
    roomService: Optional[int] = 0
    roomQuality: Optional[int] = 0
    ambience: Optional[int] = 0


class Alvie(BaseModel):
    rating: Optional[AlvieRating] = AlvieRating()
    comments: Optional[list] = []


from typing import Optional, List
from pydantic import BaseModel


class MultiText(BaseModel):
    defaultText: str = ''
    translatedTexts: Optional[list] = []


class PropertyPrimaryDetails(BaseModel):
    legalOwner: str


class PropertyGeneralInformation(BaseModel):
    roomCount: int
    floorCount: int = 0
    checkInTime: str = ''
    checkOutTime: str = ''
    hotelInfoWebsite: str
    longitude: float = None
    latitude: float = None
    baseLanguage: str = 'E'


class PropertyCurrencyFormatting(BaseModel):
    currencyCode: str = ''
    currencyFormat: str = ''
    decimalPositions: str = ''


class PropertyControls(BaseModel):
    class PropertySellControl(BaseModel):
        startDate: str = None

    class CurrencyFormatting(BaseModel):
        currencyCode: str = 'USD'
        currencyFormat: str = ''
        decimalPositions: int = 2

    class CateringCurrencyFormatting(BaseModel):
        currencyCode: str = 'USD'
        currencyFormat: str = ''

    class DateTimeFormatting(BaseModel):
        longDateFormat: str = ''
        shortDateFormat: str = ''
        timeFormat: str = ''
        timeZoneRegion: str = ''

    class ApplicationMode(BaseModel):
        mbsSupported: Optional[bool] = False

    sellControls: PropertySellControl = PropertySellControl()
    currencyFormatting: CurrencyFormatting = CurrencyFormatting()
    cateringCurrencyFormatting: CateringCurrencyFormatting = CateringCurrencyFormatting()
    dateTimeFormatting: DateTimeFormatting = DateTimeFormatting()
    applicationMode: ApplicationMode = ApplicationMode()


class PropertyCommunication(BaseModel):
    phoneNumber: str = ''
    email: str = None


class PropertyAddress(BaseModel):
    class CountryInfo(BaseModel):
        code: str = ''

    addressLine: list
    cityName: str = ''
    postalCode: str = ''
    state: str = ''
    country: CountryInfo
    regionCode = "NA"


class PropertyAttractions(BaseModel):
    name: str
    address: Optional[dict]
    operationHours: Optional[str]
    code: str


class PropertyBuilding(BaseModel):
    code: str
    description: str
    rooms: List[str]
    credits: Optional[int]


class PropertyRoom(BaseModel):
    class RoomType(BaseModel):
        pseudo: bool
        suite: bool
        roomClass: str
        houseKeeping: bool
        minimumOccupancy: Optional[int] = 0
        maximumOccupancy: Optional[int] = 0
        accessible: bool
        roomType: str
        meetingRoom: bool

    class RoomFeatures(BaseModel):
        code: str
        description: str
        orderSequence: int

    class ConnectingRoom(BaseModel):
        roomId: str

    roomType: Optional[RoomType]
    roomFeatures: List[RoomFeatures] = []
    roomDescription: str
    accessible: Optional[bool] = False
    roomId: str
    meetingRoom: Optional[bool] = False
    roomComponents: Optional[list] = []
    connectingRooms: Optional[List[ConnectingRoom]] = []
    rateAmount: Optional[dict] = {}
    maximumOccupancy: Optional[int] = 0
    sellSequence: Optional[int]
    keyOptions: Optional[list] = []
    turndownService: Optional[bool]
    roomSection: Optional[dict] = {}


class PropertyRoomClass(BaseModel):
    class RoomTypes(BaseModel):
        class RoomType(BaseModel):
            roomClass: str
            shortDescription: str
            pseudo: bool
            accessible: bool
            sendToInterface: bool
            sellSequence: int = 0
            suite: bool
            meetingRoom: bool
            roomType: str
            numberOfRooms: int
            inactive: bool

        roomTypeSummary: List[RoomType] = []

    description: MultiText = MultiText()
    sequence: int
    code: str
    inactive: bool
    roomTypes: RoomTypes


class PropertyAirport(BaseModel):
    description: str
    distance: str
    distanceType: str
    direction: str
    transportationOptions: Optional[list] = []
    code: str


class PropertyTransportation(BaseModel):
    transportationCode: str
    priceRange: str
    description: str
    relativePosition: Optional[dict] = {}
    phoneNumber: str


class PropertyCreditCardTypes(BaseModel):
    code: str
    description: MultiText = MultiText()


class PropertyHotelAmenities(BaseModel):
    description: str
    featureCode: str
    orderSequence: Optional[int] = -1
    amenityType: str
    beginDate: str


class PropertyAccountsReceivableDetails(BaseModel):
    accountName: str
    accountId: Optional[dict] = {}
    accountNo: str
    profileId: Optional[dict] = {}
    creditLimit: Optional[dict] = {}
    contactName: str
    monthEndCalcYN: bool
    address: Optional[dict] = {}
    email: Optional[dict] = {}
    status: Optional[dict]
    batchStatement: bool
    printFoliosWithStatement: bool
    emailStatementsReminders: bool
    primary: bool
    type: str
    accountTypeDescription: str
    permanent: bool


class PropertyFrontOfficeStats(BaseModel):
    code: str
    value: int


class Property(BaseModel):
    id: str
    name: str = ''
    hotelId: str = ''
    orgId: str = ''
    primaryDetails: PropertyPrimaryDetails
    generalInformation: PropertyGeneralInformation
    accommodationDetails: dict = {}
    propertyControls: PropertyControls = PropertyControls()
    communication: PropertyCommunication
    address: PropertyAddress
    hotelRateRanges: Optional[list]
    alternateHotels: Optional[list]
    hotelCorporateInformations: Optional[dict]
    # attractions: List[PropertyAttractions] = []
    buildings: Optional[List[PropertyBuilding]] = []
    rooms: List[PropertyRoom]
    roomHierarchies: Optional[list] = []
    roomClasses: Optional[List[PropertyRoomClass]] = []


if __name__ == '__main__':
    model = Property(
        id=1,
        name="happy",
        clientId='123',
        orgId='11249',
        primaryDetails={'legalOwner': 'Fam Living'},
        generalInformation={'roomCount': 10, 'hotelInfoWebsite': 'https://oracle.com'},
        communication={'phoneNumber': '1234', 'email': 'test@gmail.com'},
        address={'addressLine': [], 'cityName': 'Dubai', 'postalCode': '20232', 'state': 'Dubai',
                 'country': {'code': 'UAE'}, 'regionCode': 'AS'},
        rooms=[
            {
                'roomId': 1,
                'roomDescription': 'abcd',
                'maximumOccupancy': 2
            }
        ]
    )

    print(model.dict())



import gcsfs
import pickle
import os
from loguru import logger
import upswingutil as ul


def get_model_from_cloud_storage(file_path, file_name, token=None):
    __G_CLOUD_PROJECT__ = os.getenv('G_CLOUD_PROJECT', ul.G_CLOUD_PROJECT)
    __secret__ = os.getenv('FIREBASE', ul.FIREBASE)
    __filepath__ = f'{__G_CLOUD_PROJECT__}.appspot.com/{file_path}/{file_name}'
    try:
        if token:
            fs = gcsfs.GCSFileSystem(project=__G_CLOUD_PROJECT__, token=token)
        else:
            fs = gcsfs.GCSFileSystem(project=__G_CLOUD_PROJECT__, token=f"{__secret__}")

        with fs.open(__filepath__, 'rb') as f:
            return pickle.load(f)
    except Exception as e:
        logger.error(f'Unable to get model : {__filepath__}')
        logger.error(e)
        return None


def upload_model_to_cloud_storage(input_data, file_path, file_name, token=None):
    __G_CLOUD_PROJECT__ = os.getenv('G_CLOUD_PROJECT', ul.G_CLOUD_PROJECT)
    __secret__ = os.getenv('FIREBASE', ul.FIREBASE)
    __filepath__ = f'{__G_CLOUD_PROJECT__}.appspot.com/{file_path}/{file_name}'
    try:
        if token:
            fs = gcsfs.GCSFileSystem(project=__G_CLOUD_PROJECT__, token=token)
        else:
            fs = gcsfs.GCSFileSystem(project=__G_CLOUD_PROJECT__, token=f"{__secret__}")

        with fs.open(__filepath__, 'wb') as f:
            f.write(pickle.dumps(input_data))
            return True
    except Exception as e:
        logger.error(f'Unable to write model : {__filepath__}')
        logger.error(e)
        return False


def upload_keras_model_to_cloud_storage_test(input_data, file_path, file_name, token=None):
    __G_CLOUD_PROJECT__ = os.getenv('G_CLOUD_PROJECT', ul.G_CLOUD_PROJECT)
    __secret__ = os.getenv('FIREBASE', ul.FIREBASE)
    __filepath__ = f'{__G_CLOUD_PROJECT__}.appspot.com/{file_path}/{file_name}'
    try:
        if token:
            fs = gcsfs.GCSFileSystem(project=__G_CLOUD_PROJECT__, token=token)
        else:
            fs = gcsfs.GCSFileSystem(project=__G_CLOUD_PROJECT__, token=f"{__secret__}")

        with fs.open(__filepath__, 'wb') as f:
            f.write(input_data)
            return True
    except Exception as e:
        logger.error(f'Unable to write model : {__filepath__}')
        logger.error(e)
        return False


def get_keras_model_from_cloud_storage_test(file_path, file_name="test.h5", token=None):
    __G_CLOUD_PROJECT__ = os.getenv('G_CLOUD_PROJECT', ul.G_CLOUD_PROJECT)
    __secret__ = os.getenv('FIREBASE', ul.FIREBASE)
    __filepath__ = f'{__G_CLOUD_PROJECT__}.appspot.com/{file_path}/{file_name}'
    try:
        if token:
            fs = gcsfs.GCSFileSystem(project=__G_CLOUD_PROJECT__, token=token)
        else:
            fs = gcsfs.GCSFileSystem(project=__G_CLOUD_PROJECT__, token=f"{__secret__}")

        with fs.open(__filepath__, mode='rb') as file:  # b is important -> binary
            data = file.read()
        return data
    except Exception as e:
        logger.error(f'Unable to get model : {__filepath__}')
        logger.error(e)
        return None


if __name__ == '__main__':
    f = open("upswingutil/resource/test.h5", "rb")
    print(f.read())
    with open("upswingutil/resource/test.h5", mode='rb') as file:  # b is important -> binary
        data = file.read()
    ul.FIREBASE = 'C:/Upswing/Projects/Code/agent-oracle/SECRET/aura-staging-31cae-firebase-adminsdk-dyolr-7c135838e9.json'
    ul.G_CLOUD_PROJECT = 'aura-staging-31cae'
    # data = f.read()
    # upload_model_to_cloud_storage_test(data, '11281', 'test.h5')
    get_keras_model_from_cloud_storage_test('11281')
    # data = 'test'

    # print(upload_model_to_cloud_storage(data, 'OHIPSB', 'test.txt'))
    # output = get_model_from_cloud_storage('OHIPSB', 'test.txt')
    # print(output)



from mongoengine import *


class ProfileIDObject(DynamicEmbeddedDocument):
    profile = StringField()
    corporateId = StringField()


class GuestsModel(DynamicDocument):
    meta = {'collection': 'guests'}

    _id = StringField(primary_key=True)
    birthDate = StringField()
    addresses = DictField()
    comments = DictField()
    createDateTime = StringField()
    creatorId = StringField()
    businessSegments = StringField()
    customer = DictField()
    company = DictField()
    emails = DictField()
    firstName = StringField()
    idObj = EmbeddedDocumentField(ProfileIDObject)
    keywords = ListField()
    lastModifierId = StringField()
    lastModifyDateTime = StringField()
    lastName = StringField()
    lastStayInfo = DictField()
    mailingActions = DictField()
    markForHistory = BooleanField()
    middleName = StringField()
    preferenceCollection = DictField()
    privacyInfo = DictField()
    profileAccessType = DictField()
    profileDeliveryMethods = DictField()
    profileIndicators = ListField(DictField())
    profileMemberships = DictField()
    profileRestrictions = DictField()
    profileType = StringField()
    registeredProperty = StringField()
    relationships = DictField()
    relationshipsSummary = DictField()
    salesInfo = DictField()
    statusCode = StringField()
    stayReservationInfoList = DictField()
    subscriptions = ListField(DictField())
    taxInfo = DictField()
    telephones = DictField()


class GuestsModelV2(DynamicDocument):
    meta = {'collection': 'guests'}

    _id = StringField(primary_key=True)
    idObj = EmbeddedDocumentField(ProfileIDObject)
    profileType = StringField()
    title = StringField()
    firstName = StringField()
    lastName = StringField()
    datesAndDurations = DictField()
    address = DictField()
    contactInfo = DictField()
    privacyInfo = DictField()
    documents = DictField()
    metaInfo = DictField()
    company = DictField()
    registeredProperty = StringField()
    auraRecordUpdatedOn = DateTimeField()



from typing import Optional
from typing import List
from upswingutil.schema import Token, Alvie
from pydantic import BaseModel


class PropertyInfo(BaseModel):
    hotelId: str
    clientId: str
    orgId: str
    agent: str
    hotelName: str
    numberOfRooms: int
    roomId: str
    roomName: str
    roomType: str
    roomClass: str
    roomTypeCharged: str
    roomNumberLocked: bool
    pseudoRoom: bool


class TravelType(BaseModel):
    id: int
    name: str
    type: str
    score: float
    addedBy: str


class GuestProfile(BaseModel):
    guestId: str
    firstName: Optional[str] = ''
    middleName: Optional[str] = ''
    lastName: Optional[str] = ''
    primary: Optional[bool] = False
    address: Optional[dict]
    profileType: Optional[str] = ''
    primary: Optional[bool] = False
    arrivalTransport: Optional[dict]
    departureTransport: Optional[dict]


class GuestInfo(BaseModel):
    adultCount: int
    childrenCount: int
    infantCount: int
    guest_list: List[GuestProfile]


class BookingChannelInfo(BaseModel):
    saleSourceType: str
    saleSourceCode: str
    sourceCode: str
    sourceDescription: str
    channelCode: str
    bookingMedium: Optional[str] = ''
    bookingMediumDescription: Optional[str] = ''


class BookingFinancialInfo(BaseModel):
    class RoomRate(BaseModel):
        start: str
        end: str
        total: dict
        rates: dict
        guestCounts: dict
        suppressRate: bool = False
        houseUseOnly: bool = False
        complimentary: bool = False
        discountAllowed: bool = False
        bogoDiscount: bool = False

    currencyCode: str
    rateAmount: float
    paymentMethod: str
    fixedRate: Optional[bool]
    rateSuppressed: Optional[bool]
    ratePlanCode: Optional[str]
    total: Optional[dict]
    roomRates: Optional[List[RoomRate]] = []


class Reservation(BaseModel):
    id: str
    idObj: dict
    orgId: str
    agent: str
    hotelId: str
    hotelName: str
    globalId: str
    status: str
    alerts: list
    financeInfo: dict
    guestInfo: dict
    arrivalDate: str
    departureDate: str
    expectedTimes: dict
    bookingInfo: dict
    token: Optional[Token] = None
    roomStay: Optional[dict] = None
    originalTimeSpan: Optional[dict] = None
    callHistory: Optional[list] = []
    cancellation: Optional[list] = []
    comments: Optional[list] = []
    createBusinessDate: Optional[str] = ''
    createDateTime: Optional[str] = ''
    guestLocators: Optional[dict]
    policies: Optional[dict] = None
    inventoryItems: Optional[dict] = None
    linkedReservation: Optional[dict] = None
    preferences: Optional[list] = []
    metaInfo: dict
    daily_activity: Optional[list] = []
    eCertificates: Optional[list] = []
    memberships: Optional[list] = []
    packages: Optional[list] = []
    folioInformation: Optional[dict] = None
    alvie: Optional[Alvie] = Alvie()


if __name__ == '__main__':
    reservation = Reservation(

    )
    # g = GuestProfile(
    #     guestId=123,
    #     firstName=None
    # )
    # print(g.json())
